# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest

from xcube_resampling.gridmapping import assertions


class Testassertions(unittest.TestCase):
    # --- assert_given ---
    def test_assert_given_ok(self):
        assertions.assert_given("hello")

    def test_assert_given_fail(self):
        with self.assertRaises(ValueError) as cm:
            assertions.assert_given("", name="arg")
        self.assertIn("arg must be given", str(cm.exception))

    # --- assert_instance ---
    def test_assert_instance_ok(self):
        assertions.assert_instance(5, int)

    def test_assert_instance_fail(self):
        with self.assertRaises(TypeError) as cm:
            assertions.assert_instance("s", int, name="val")
        self.assertIn("val must be an instance of", str(cm.exception))

    def test_assert_instance_tuple_dtype(self):
        assertions.assert_instance(5, (int, float))  # should pass

    # --- assert_in ---
    def test_assert_in_ok(self):
        assertions.assert_in(1, [1, 2, 3])

    def test_assert_in_fail(self):
        with self.assertRaises(ValueError) as cm:
            assertions.assert_in("z", ["a", "b"], name="char")
        self.assertIn("char must be one of", str(cm.exception))

    # --- assert_true ---
    def test_assert_true_ok(self):
        assertions.assert_true(True, "must be true")

    def test_assert_true_fail(self):
        with self.assertRaises(ValueError) as cm:
            assertions.assert_true(False, "bad value")
        self.assertEqual(str(cm.exception), "bad value")



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from collections.abc import Hashable
from typing import Iterable, Literal, Mapping, Sequence

import numpy as np
import pandas as pd
import xarray as xr

from .constants import (
    LOG,
    TemporalAggMethod,
    TemporalAggMethods,
    TemporalInterpMethod,
    TemporalInterpMethods,
)
from .utils import _select_variables


def resample_in_time(
    source_ds: xr.Dataset,
    frequency: str,
    *,
    variables: str | Iterable[str] | None = None,
    interp_methods: TemporalInterpMethods | None = None,
    agg_methods: TemporalAggMethods | None = None,
    offset: str | None = None,
    tolerance: str | None = None,
) -> xr.Dataset:
    """
    Resample a dataset along the time dimension.

    This function allows both **upsampling** (interpolation) and **downsampling**
    (aggregation) of temporal data variables. It wraps xarray's
    [`resample`](https://docs.xarray.dev/en/stable/generated/xarray.DataArray.resample.html)
    functionality and automatically determines whether to apply interpolation or
    aggregation based on the input frequency and dataset time resolution.

    Args:
        source_ds: Input xarray Dataset containing a `time` dimension.
        frequency: Target temporal frequency, following
            [Pandas offset aliases](https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases).
            Format `<count><offset>`, where `<offset>`
            may be one of 's', 'min', 'h', 'D', 'W', 'MS', 'ME', 'QS', 'QE' 'YS', 'YE'.
        variables: Optional. Names of variables to resample. If None, all
            data variables are processed, which have a time coordinate.
        interp_methods: Optional interpolation method(s) for upsampling. Can be:

            - A single method for all variables
            - A list of methods (applied sequentially, each output saved as
              `varname_method`)
            - A dictionary mapping variable names or dtypes to method(s)

            Available methods are defined in
            [`TemporalInterpMethod`](api.md/#xcube_resampling.constants.TemporalInterpMethod).
            Default is "nearest" for integer data, otherwise "linear".
        agg_methods: Optional aggregation method(s) for downsampling. Can be:

            - A single method for all variables
            - A list of methods (applied sequentially, each output saved as
              `varname_method`)
            - A dictionary mapping variable names or dtypes to method(s)

            Available methods are defined in
            [`TemporalAggMethod`](api.md/#xcube_resampling.constants.TemporalAggMethod).
            The placeholder `'percentile_<p>'` can be used for percentiles
            (e.g., `'percentile_90'`). Default is `'nearest'` for integer data,
            otherwise `'mean'`.
        offset: Optional offset to adjust resampled time labels. Uses the
            same syntax as frequency.
        tolerance: Optional maximum allowed distance for selective downsampling
            methods (e.g., `'backfill'`, `'ffill'`, `'nearest'`), following
            [Pandas period aliases](https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#period-aliases).
            Format `<count><period>`. Defaults to None.

    Returns:
        A new xarray Dataset resampled along the time dimension.

    Notes:
        - The function automatically decides whether to apply interpolation or
          aggregation based on the target frequency and the dataset’s temporal
          resolution, if `agg_methods` or `interp_methods` are not provided.
        - If the time series is highly irregular (coefficient of variation > 5%),
          explicit `agg_methods` or `interp_methods` must be provided.
    """

    if "time" not in source_ds.dims:
        raise ValueError("Dataset must have a 'time' dimension.")

    if variables:
        source_ds = _select_variables(source_ds, variables)

    if frequency == "all":
        days = int((source_ds.time[-1] - source_ds.time[0]) / np.timedelta64(1, "D"))
        frequency = f"{days + 1}D"

    guessed_operation = _guess_resampling_operation(
        source_ds, frequency, interp_methods, agg_methods
    )

    if guessed_operation == "agg":
        target_ds = _apply_aggregation(
            source_ds,
            frequency,
            agg_methods=agg_methods,
            offset=offset,
            tolerance=tolerance,
        )

    elif guessed_operation == "interp":
        target_ds = _apply_interpolation(
            source_ds,
            frequency,
            interp_methods=interp_methods,
            offset=offset,
        )

    else:
        LOG.warning(
            "Could not determine resampling operation. Please pass agg_methods "
            "or interp_methods. Returning original dataset as is."
        )
        return source_ds

    time_coverage_start = pd.Timestamp(source_ds.time[0].item()).isoformat()
    time_coverage_end = pd.Timestamp(source_ds.time[-1].item()).isoformat()

    target_ds.attrs.update(
        time_coverage_start=time_coverage_start,
        time_coverage_end=time_coverage_end,
    )

    return target_ds


def _apply_aggregation(
    dataset: xr.Dataset,
    frequency: str,
    agg_methods: TemporalAggMethods | None = None,
    offset: str | None = None,
    tolerance: str | None = None,
) -> xr.Dataset:
    percentile_prefix = "percentile_"

    resampled_dataset = xr.Dataset(attrs=dataset.attrs)
    for var_name, data_array in dataset.data_vars.items():
        if "time" not in data_array.coords:
            continue
        var_methods = _get_temporal_agg_method(agg_methods, var_name, data_array)
        resampler = data_array.resample(
            skipna=True, closed="left", label="left", time=frequency, offset=offset
        )
        for method in var_methods:
            method_args, method_kwargs = [], {}
            method_postfix = method
            if method.startswith(percentile_prefix):
                p = int(method[len(percentile_prefix) :])
                method_args = [p / 100.0]
                method_postfix = f"p{p}"
                method = "quantile"
            method_kwargs = _get_agg_method_kwargs(method, tolerance)
            func = getattr(resampler, method)
            if agg_methods is None or isinstance(agg_methods, str):
                var_name_out = var_name
            else:
                var_name_out = f"{var_name}_{method_postfix}"
            resampled_dataset[var_name_out] = func(*method_args, **method_kwargs)

    return resampled_dataset


def _apply_interpolation(
    dataset: xr.Dataset,
    frequency: str,
    interp_methods: TemporalInterpMethods | None = None,
    offset: str | None = None,
) -> xr.Dataset:
    resampled_dataset = xr.Dataset(attrs=dataset.attrs)
    for var_name, data_array in dataset.data_vars.items():
        if "time" not in data_array.coords:
            continue
        var_methods = _get_temporal_interp_method(interp_methods, var_name, data_array)
        resampler = data_array.resample(
            skipna=True, closed="left", label="left", time=frequency, offset=offset
        )
        for method in var_methods:
            func = getattr(resampler, "interpolate")
            if interp_methods is None or isinstance(interp_methods, str):
                var_name_out = var_name
            else:
                var_name_out = f"{var_name}_{method}"
            resampled_dataset[var_name_out] = func(kind=method)
    return resampled_dataset


def _get_agg_method_kwargs(
    agg_method: str,
    tolerance: str | None = None,
) -> dict:
    if agg_method in {"backfill", "bfill", "ffill", "nearest", "pad"}:
        kwargs = {"tolerance": tolerance}
    elif agg_method in {"all", "any", "count"}:
        kwargs = {"dim": "time", "keep_attrs": True}
    elif agg_method in {
        "cumprod",
        "cumsum",
        "first",
        "last",
        "max",
        "min",
        "mean",
        "median",
        "prod",
        "quantile",
        "std",
        "sum",
        "var",
    }:
        kwargs = {"dim": "time", "keep_attrs": True, "skipna": True}
    else:
        raise ValueError(f"Aggregation method {agg_method!r} not supported.")
    return kwargs


def _guess_resampling_operation(
    ds: xr.Dataset,
    frequency: str,
    interp_methods: TemporalInterpMethods | None = None,
    agg_methods: TemporalAggMethods | None = None,
) -> Literal["agg", "interp", None]:

    if agg_methods and interp_methods:
        raise ValueError(
            "Please provide either agg_methods or " "interp_methods, not both."
        )
    elif agg_methods:
        return "agg"
    elif interp_methods:
        return "interp"

    time = ds["time"].values
    if len(time) < 2:
        raise ValueError("Not enough time points to resample.")
    deltas = np.diff(time).astype("timedelta64[ns]").astype(float)
    mean_delta = np.mean(deltas)
    std_delta = np.std(deltas)
    if mean_delta == 0 or (std_delta / mean_delta) > 0.05:
        # irregular time delta in the time series
        return None

    mean_td = pd.to_timedelta(int(mean_delta), unit="ns")
    target_td = pd.Timedelta(frequency)
    ratio = mean_td / target_td
    if ratio < 1:
        return "agg"
    elif ratio > 1:
        return "interp"
    else:
        return None


def _get_temporal_interp_method(
    interp_methods: TemporalInterpMethods | None,
    key: Hashable,
    var: xr.DataArray,
) -> TemporalInterpMethod | Sequence[TemporalInterpMethod]:
    def assign_defaults(data_type: np.dtype) -> TemporalInterpMethod:
        return "nearest" if np.issubdtype(data_type, np.integer) else "linear"

    if isinstance(interp_methods, Mapping):
        interp_method = interp_methods.get(str(key), interp_methods.get(var.dtype))
        if interp_method is None:
            LOG.warning(
                f"Interpolation method could not be derived from the mapping "
                f"`interp_methods` for data variable {key!r} with data type "
                f"{var.dtype!r}. Defaults are assigned."
            )
            interp_method = assign_defaults(var.dtype)
    elif isinstance(interp_methods, str) or isinstance(interp_methods, Sequence):
        interp_method = interp_methods
    else:
        interp_method = assign_defaults(var.dtype)

    if isinstance(interp_method, str):
        interp_method = [interp_method]
    return interp_method


def _get_temporal_agg_method(
    agg_methods: TemporalAggMethods | None,
    key: Hashable,
    var: xr.DataArray,
) -> TemporalAggMethod | Sequence[TemporalAggMethod]:
    def assign_defaults(data_type: np.dtype) -> TemporalAggMethod:
        return "nearest" if np.issubdtype(data_type, np.integer) else "mean"

    if isinstance(agg_methods, Mapping):
        agg_method = agg_methods.get(str(key), agg_methods.get(var.dtype))
        if agg_method is None:
            LOG.warning(
                f"Aggregation method could not be derived from the mapping "
                f"`agg_methods` for data variable {key!r} with data type "
                f"{var.dtype!r}. Defaults are assigned."
            )
            agg_method = assign_defaults(var.dtype)
    elif isinstance(agg_methods, str) or isinstance(agg_methods, Sequence):
        agg_method = agg_methods
    else:
        agg_method = assign_defaults(var.dtype)

    if isinstance(agg_method, str):
        agg_method = [agg_method]
    return agg_method



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import dask.array as da
import numpy as np
import pandas as pd
import pyproj
import xarray as xr


def create_2x2_dataset_with_irregular_coords():
    lon = np.array([[1.0, 6.0], [0.0, 2.0]])
    lat = np.array([[56.0, 53.0], [52.0, 50.0]])
    rad = np.array([[1.0, 2.0], [3.0, 4.0]])
    return xr.Dataset(
        dict(rad=xr.DataArray(rad, dims=("y", "x"))),
        coords=dict(
            lon=xr.DataArray(lon, dims=("y", "x")),
            lat=xr.DataArray(lat, dims=("y", "x")),
        ),
    )


def create_2x2x2_dataset_with_irregular_coords():
    lon = np.array([[1.0, 6.0], [0.0, 2.0]])
    lat = np.array([[56.0, 53.0], [52.0, 50.0]])
    time = pd.date_range("2025-08-01", periods=2)
    rad = np.array([[[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]]])
    return xr.Dataset(
        dict(
            rad=xr.DataArray(rad, dims=("time", "y", "x")),
            time_series=xr.DataArray(np.array([1, 2]), dims=("time")),
        ),
        coords=dict(
            lon=xr.DataArray(lon, dims=("y", "x")),
            lat=xr.DataArray(lat, dims=("y", "x")),
            time=time,
        ),
    )


def create_8x6_dataset_with_regular_coords():
    res = 0.1
    return xr.Dataset(
        data_vars=dict(
            refl=xr.DataArray(
                np.array(
                    [
                        [0, 1, 0, 2, 0, 3, 0, 4],
                        [2, 0, 3, 0, 4, 0, 1, 0],
                        [0, 4, 0, np.nan, 0, 2, 0, 3],
                        [1, 0, 2, 0, 3, 0, 4, 0],
                        [0, 3, 0, 4, 0, 1, 0, 2],
                        [4, 0, 1, 0, 2, 0, 3, 0],
                    ],
                    dtype=np.float64,
                ),
                dims=("lat", "lon"),
            ),
            ndvi=xr.DataArray(
                np.array(
                    [
                        [0, 1, 0, 2, 0, 3, 0, 4],
                        [2, 0, 3, 0, 4, 0, 1, 0],
                        [0, 4, 0, np.nan, 0, 2, 0, 3],
                        [1, 0, 2, 0, 3, 0, 4, 0],
                        [0, 3, 0, 4, 0, 1, 0, 2],
                        [4, 0, 1, 0, 2, 0, 3, 0],
                    ],
                    dtype=np.float64,
                ),
                dims=("lat", "lon"),
            ),
        ),
        coords=dict(
            lon=xr.DataArray(50.0 + res * np.arange(0, 8) + 0.5 * res, dims="lon"),
            lat=xr.DataArray(10.6 - res * np.arange(0, 6) - 0.5 * res, dims="lat"),
        ),
    )


def create_2x8x6_dataset_with_regular_coords():
    ds = create_8x6_dataset_with_regular_coords()
    array_3d = np.repeat(ds.refl.values[np.newaxis, :, :], 2, axis=0)
    time = pd.date_range("2025-08-01", periods=2)
    ds_3d = xr.Dataset(coords=dict(time=time, lat=ds.lat, lon=ds.lon))
    ds_3d["refl"] = (("time", "lat", "lon"), array_3d)
    return ds_3d


def create_5x5_dataset_regular_utm():
    x = np.arange(565300.0, 565800.0, 100.0)
    y = np.arange(5934300.0, 5933800.0, -100.0)
    spatial_ref = np.array(0)
    band_1 = np.arange(25).reshape((5, 5))
    ds = xr.Dataset(
        dict(
            band_1=xr.DataArray(
                band_1, dims=("y", "x"), attrs=dict(grid_mapping="spatial_ref")
            )
        ),
        coords=dict(x=x, y=y, spatial_ref=spatial_ref),
    )
    ds.spatial_ref.attrs = pyproj.CRS.from_epsg("32632").to_cf()
    return ds


def create_5x5_dataset_regular_utm_antimeridian():
    # Regular 5x5 grid in meters (UTM zone 60N, close to 180°E)
    x = np.arange(828800.0, 829300.0, 100.0)
    y = np.arange(1101700.0, 1101200.0, -100.0)

    spatial_ref = np.array(0)
    band_1 = np.arange(25).reshape((5, 5))

    ds = xr.Dataset(
        dict(
            band_1=xr.DataArray(
                band_1,
                dims=("y", "x"),
                attrs=dict(grid_mapping="spatial_ref"),
            )
        ),
        coords=dict(
            x=x,
            y=y,
            spatial_ref=spatial_ref,
        ),
    )

    # Attach UTM zone 60N CRS
    ds.spatial_ref.attrs = pyproj.CRS.from_epsg("32660").to_cf()

    return ds


def create_2x5x5_dataset_regular_utm():
    x = np.arange(565300.0, 565800.0, 100.0)
    y = np.arange(5934300.0, 5933800.0, -100.0)
    time = pd.date_range("2025-08-01", periods=2)
    spatial_ref = np.array(0)
    band_1 = np.arange(25).reshape((5, 5))
    band_1 = np.repeat(band_1[np.newaxis, :, :], 2, axis=0)
    ds = xr.Dataset(
        dict(
            band_1=xr.DataArray(
                band_1, dims=("time", "y", "x"), attrs=dict(grid_mapping="spatial_ref")
            )
        ),
        coords=dict(time=time, x=x, y=y, spatial_ref=spatial_ref),
    )
    ds.spatial_ref.attrs = pyproj.CRS.from_epsg("32632").to_cf()
    return ds


def create_large_dataset_for_reproject():
    nt, nx, ny = 10, 100, 100
    chunks = {"time": 2, "x": 25, "y": 25}

    # Create coordinate arrays
    times = pd.date_range("2023-01-01", periods=nt, freq="D")
    x = np.linspace(3900000, 4500000, nx)
    y = np.linspace(2600000, 3200000, ny)
    temp_data = da.from_array(
        np.arange(nt * nx * ny, dtype=np.float32).reshape(nt, nx, ny),
        chunks=(chunks["time"], chunks["x"], chunks["y"]),
    )
    onedim_data = da.from_array(np.arange(nt), chunks=(chunks["time"]))
    spatial_ref = np.array(0)
    ds = xr.Dataset(
        dict(
            temperature=xr.DataArray(
                temp_data,
                dims=("time", "y", "x"),
                attrs=dict(grid_mapping="spatial_ref"),
            ),
            onedim_data=xr.DataArray(onedim_data, dims="time"),
        ),
        coords=dict(time=times, x=x, y=y, spatial_ref=spatial_ref),
    )
    ds.spatial_ref.attrs = pyproj.CRS.from_epsg("3035").to_cf()
    return ds


def create_2x2_dataset_with_irregular_coords_antimeridian():
    lon = np.array([[+179.0, -176.0], [+178.0, +180.0]])
    lat = np.array([[56.0, 53.0], [52.0, 50.0]])
    rad = np.array([[1.0, 2.0], [3.0, 4.0]])
    return xr.Dataset(
        dict(
            rad=xr.DataArray(rad, dims=("y", "x")),
        ),
        coords=dict(
            lon=xr.DataArray(lon, dims=("y", "x")),
            lat=xr.DataArray(lat, dims=("y", "x")),
        ),
    )


def create_4x4_dataset_with_irregular_coords():
    lon = np.array(
        [
            [1.0, 2.0, 3.0, 4.0],
            [0.0, 1.0, 2.0, 3.0],
            [-1.0, 0.0, 1.0, 2.0],
            [-2.0, -1.0, 0.0, 1.0],
        ]
    )
    lat = np.array(
        [
            [56.0, 55.0, 54.0, 53.0],
            [55.0, 54.0, 53.0, 52.0],
            [54.0, 53.0, 52.0, 51.0],
            [53.0, 52.0, 51.0, 50.0],
        ]
    )
    rad = np.array(
        [
            [1.0, 2.0, 3.0, 4.0],
            [5.0, 6.0, 7.0, 8.0],
            [9.0, 10.0, 11.0, 12.0],
            [13.0, 14.0, 15.0, 16.0],
        ]
    )
    return xr.Dataset(
        dict(
            rad=xr.DataArray(rad, dims=("y", "x")),
        ),
        coords=dict(
            lon=xr.DataArray(lon, dims=("y", "x")),
            lat=xr.DataArray(lat, dims=("y", "x")),
        ),
    )


def create_2x4x4_dataset_with_irregular_coords():
    ds = create_4x4_dataset_with_irregular_coords()
    array_3d = np.repeat(ds.rad.values[np.newaxis, :, :], 2, axis=0)
    time = pd.date_range("2025-08-01", periods=2)
    ds_3d = xr.Dataset(coords=dict(time=time, lat=ds.lat, lon=ds.lon))
    ds_3d["rad"] = (("time", "lat", "lon"), array_3d)
    return ds_3d


def create_s2plus_dataset():
    x = xr.DataArray(
        [310005.0, 310015.0, 310025.0, 310035.0, 310045.0],
        dims=["x"],
        attrs=dict(units="m", standard_name="projection_x_coordinate"),
    )
    y = xr.DataArray(
        [5689995.0, 5689985.0, 5689975.0, 5689965.0, 5689955.0],
        dims=["y"],
        attrs=dict(units="m", standard_name="projection_y_coordinate"),
    )
    lon = xr.DataArray(
        [
            [0.272763, 0.272906, 0.273050, 0.273193, 0.273336],
            [0.272768, 0.272911, 0.273055, 0.273198, 0.273342],
            [0.272773, 0.272917, 0.273060, 0.273204, 0.273347],
            [0.272779, 0.272922, 0.273066, 0.273209, 0.273352],
            [0.272784, 0.272927, 0.273071, 0.273214, 0.273358],
        ],
        dims=["y", "x"],
        attrs=dict(units="degrees_east", standard_name="longitude"),
    )
    lat = xr.DataArray(
        [
            [51.329464, 51.329464, 51.329468, 51.32947, 51.329475],
            [51.329372, 51.329376, 51.32938, 51.329384, 51.329388],
            [51.329285, 51.329285, 51.32929, 51.329292, 51.329296],
            [51.329193, 51.329197, 51.32920, 51.329205, 51.329205],
            [51.329100, 51.329105, 51.32911, 51.329113, 51.329117],
        ],
        dims=["y", "x"],
        attrs=dict(units="degrees_north", standard_name="latitude"),
    )
    rrs_443 = xr.DataArray(
        [
            [0.014000, 0.014000, 0.016998, 0.016998, 0.016998],
            [0.014000, 0.014000, 0.016998, 0.016998, 0.016998],
            [0.019001, 0.019001, 0.016998, 0.016998, 0.016998],
            [0.019001, 0.019001, 0.016998, 0.016998, 0.016998],
            [0.019001, 0.019001, 0.016998, 0.016998, 0.016998],
        ],
        dims=["y", "x"],
        attrs=dict(units="sr-1", grid_mapping="transverse_mercator"),
    )
    rrs_665 = xr.DataArray(
        [
            [0.025002, 0.019001, 0.008999, 0.012001, 0.022999],
            [0.028000, 0.021000, 0.009998, 0.008999, 0.022999],
            [0.036999, 0.022999, 0.007999, 0.008999, 0.023998],
            [0.041000, 0.022999, 0.007000, 0.009998, 0.021000],
            [0.033001, 0.018002, 0.007999, 0.008999, 0.021000],
        ],
        dims=["y", "x"],
        attrs=dict(units="sr-1", grid_mapping="transverse_mercator"),
    )
    transverse_mercator = xr.DataArray(
        np.array([0xFFFFFFFF], dtype=np.uint32),
        attrs=dict(
            grid_mapping_name="transverse_mercator",
            scale_factor_at_central_meridian=0.9996,
            longitude_of_central_meridian=3.0,
            latitude_of_projection_origin=0.0,
            false_easting=500000.0,
            false_northing=0.0,
            semi_major_axis=6378137.0,
            inverse_flattening=298.257223563,
        ),
    )
    return xr.Dataset(
        dict(rrs_443=rrs_443, rrs_665=rrs_665, transverse_mercator=transverse_mercator),
        coords=dict(x=x, y=y, lon=lon, lat=lat),
        attrs={
            "title": "T31UCS_20180802T105621",
            "conventions": "CF-1.6",
            "institution": "VITO",
            "product_type": "DCS4COP Sentinel2 Product",
            "origin": "Copernicus Sentinel Data",
            "project": "DCS4COP",
            "time_coverage_start": "2018-08-02T10:59:38.888000Z",
            "time_coverage_end": "2018-08-02T10:59:38.888000Z",
        },
    )


def create_highroc_dataset(no_spectra=False):
    """
    Simulates a HIGHROC OLCI L2 product in NetCDF 4 format
    """
    lon = np.array(
        [[8, 9.3, 10.6, 11.9], [8, 9.2, 10.4, 11.6], [8, 9.1, 10.2, 11.3]],
        dtype=np.float32,
    )
    lat = np.array(
        [[56, 56.1, 56.2, 56.3], [55, 55.2, 55.4, 55.6], [54, 54.3, 54.6, 54.9]],
        dtype=np.float32,
    )

    if not no_spectra:
        wavelengths = [
            (1, 400.0),
            (2, 412.5),
            (3, 442.5),
            (4, 490.0),
            (5, 510.0),
            (6, 560.0),
            (7, 620.0),
            (8, 665.0),
            (9, 673.75),
            (10, 681.25),
            (11, 708.75),
            (12, 753.75),
            (16, 778.75),
            (17, 865.0),
            (18, 885.0),
            (21, 940.0),
        ]
        rtoa_desc = "Top-of-atmosphere reflectance"
        rrs_desc = (
            "Atmospherically corrected angular dependent remote sensing reflectances"
        )
        rtoa_vars = {
            f"rtoa_{i}": create_waveband(i, wl, "1", rtoa_desc) for i, wl in wavelengths
        }
        rrs_vars = {
            f"rrs_{i}": create_waveband(i, wl, "sr^-1", rrs_desc)
            for i, wl in wavelengths
        }
    else:
        rtoa_vars = {}
        rrs_vars = {}

    return xr.Dataset(
        data_vars=dict(
            conc_chl=create_conc_chl(),
            c2rcc_flags=create_c2rcc_flag_var(),
            lon=(
                ("y", "x"),
                lon,
                dict(
                    long_name="longitude",
                    units="degrees_east",
                ),
            ),
            lat=(
                ("y", "x"),
                lat,
                dict(
                    long_name="latitude",
                    units="degrees_north",
                ),
            ),
            **rtoa_vars,
            **rrs_vars,
        ),
        attrs=dict(
            start_date="14-APR-2017 10:27:50.183264",
            stop_date="14-APR-2017 10:31:42.736226",
        ),
    )


def create_waveband(index, wavelength, units, long_name=None):
    data = np.array(
        [[7, 11, np.nan, 5], [5, 10, 2, 21], [16, 6, 20, 17]], dtype=np.float32
    )
    return (
        ("y", "x"),
        data,
        dict(
            long_name=long_name,
            units=units,
            spectral_band_index=index,
            wavelength=wavelength,
            bandwidth=15.0,
            valid_pixel_expression="c2rcc_flags.F1",
            _FillValue=np.nan,
        ),
    )


def create_conc_chl():
    data = np.array(
        [[7, 11, np.nan, 5], [5, 10, 2, 21], [16, 6, 20, 17]], dtype=np.float32
    )
    return (
        ("y", "x"),
        data,
        dict(
            long_name="Chlorophyll concentration",
            units="mg m^-3",
            _FillValue=np.nan,
            valid_pixel_expression="c2rcc_flags.F1",
        ),
    )


def create_c2rcc_flag_var():
    data = np.array([[1, 1, 1, 1], [1, 4, 1, 2], [8, 1, 1, 1]], dtype=np.uint32)
    return xr.DataArray(
        data,
        dims=("y", "x"),
        name="c2rcc_flags",
        attrs=dict(
            long_name="C2RCC quality flags",
            _Unsigned="true",
            flag_meanings="F1 F2 F3 F4",
            flag_masks=np.array([1, 2, 4, 8], np.int32),
            flag_coding_name="c2rcc_flags",
            flag_descriptions="D1 D2 D3 D4",
        ),
    )


def create_cmems_sst_flag_var():
    sea = 1
    land = 2
    lake = 4
    ice = 8
    data = np.array(
        [
            [
                [sea + ice, land + ice, lake + ice, lake],
                [sea + ice, sea, land, land],
                [sea, sea, sea, land],
            ]
        ],
        dtype=np.float32,
    )
    return xr.DataArray(
        data,
        dims=("time", "lat", "lon"),
        name="mask",
        attrs=dict(
            long_name="land sea ice lake bit mask",
            flag_masks="0b, 1b, 2b, 3b",
            flag_meanings="sea land lake ice",
            valid_min=0,
            valid_max=12,
        ),
    )


def create_cci_lccs_class_var(flag_values_as_list=False):
    data = np.array([[[30, 130, 40], [81, 201, 40], [190, 90, 50]]], dtype=np.uint8)
    var = xr.DataArray(
        data,
        dims=("time", "lat", "lon"),
        name="lccs_class",
        attrs={
            "ancillary_variables": (
                "processed_flag current_pixel_state observation_count change_count"
            ),
            # Note, number of colors in flag_colors is 37,
            # but values in flag_values is 38!
            "flag_colors": (
                "#ffff64 #ffff64 #ffff00 #aaf0f0 #dcf064 #c8c864 #006400"
                " #00a000 #00a000 #aac800 #003c00 #003c00 #005000 #285000"
                " #285000 #286400 #788200 #8ca000 #be9600 #966400 #966400"
                " #966400 #ffb432 #ffdcd2 #ffebaf #ffc864 #ffd278 #ffebaf"
                " #00785a #009678 #00dc82 #c31400 #fff5d7 #dcdcdc #fff5d7"
                " #0046c8 #ffffff"
            ),
            "flag_meanings": (
                "no_data cropland_rainfed cropland_rainfed_herbaceous_cover"
                " cropland_rainfed_tree_or_shrub_cover cropland_irrigated"
                " mosaic_cropland mosaic_natural_vegetation"
                " tree_broadleaved_evergreen_closed_to_open"
                " tree_broadleaved_deciduous_closed_to_open"
                " tree_broadleaved_deciduous_closed tree_broadleaved_deciduous_open"
                " tree_needleleaved_evergreen_closed_to_open"
                " tree_needleleaved_evergreen_closed tree_needleleaved_evergreen_open"
                " tree_needleleaved_deciduous_closed_to_open"
                " tree_needleleaved_deciduous_closed tree_needleleaved_deciduous_open"
                " tree_mixed mosaic_tree_and_shrub mosaic_herbaceous shrubland"
                " shrubland_evergreen shrubland_deciduous grassland lichens_and_mosses"
                " sparse_vegetation sparse_tree sparse_shrub sparse_herbaceous"
                " tree_cover_flooded_fresh_or_brakish_water"
                " tree_cover_flooded_saline_water shrub_or_herbaceous_cover_flooded"
                " urban bare_areas bare_areas_consolidated bare_areas_unconsolidated"
                " water snow_and_ice"
            ),
            "flag_values": (
                "0, 10, 11, 12, 20, 30, 40, 50, 60, 61, 62, 70, 71, 72, 80, 81, 82,"
                " 90, 100, 110, 120, 121, 122, 130, 140, 150, 151, 152, 153, 160, 170,"
                " 180, 190, 200, 201, 202, 210, 220"
            ),
            "long_name": "Land cover class defined in LCCS",
            "standard_name": "land_cover_lccs",
            "valid_max": 220,
            "valid_min": 1,
        },
    )
    if flag_values_as_list:
        flag_values = var.attrs["flag_values"]
        var.attrs["flag_values"] = list(map(int, flag_values.split(", ")))
    return var


def create_nx8x6_dataset_with_regular_coords(days: int):
    ds = create_8x6_dataset_with_regular_coords()
    refl_base = ds.refl.values.copy()
    ndvi_base = ds.ndvi.values.copy()
    nlat, nlon = refl_base.shape
    day_range = np.arange(0, days + 1)
    amplitude = np.arange(-days / 2, days / 2)
    time = pd.date_range("2025-08-01", periods=days)

    refl_array_3d = np.zeros((days, nlat, nlon), dtype=float)
    ndvi_array_3d = np.zeros((days, nlat, nlon), dtype=float)

    for d, a in zip(day_range, amplitude):
        refl_array_3d[d] = refl_base + a

    for d, a in zip(day_range, amplitude):
        ndvi_array_3d[d] = ndvi_base + a

    ds_3d = xr.Dataset(
        data_vars=dict(
            refl=(("time", "lat", "lon"), refl_array_3d),
            ndvi=(("time", "lat", "lon"), ndvi_array_3d),
        ),
        coords=dict(time=time, lat=ds.lat, lon=ds.lon),
    )

    return ds_3d



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from collections.abc import Container
from typing import Any

_DEFAULT_NAME = "value"


def assert_given(
    value: Any, name: str = None, exception_type: type[Exception] = ValueError
):
    """Assert *value* is not False when converted into a Boolean value.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test.
        name: Name of a variable that holds *value*.
        exception_type: The exception type. Default is ``ValueError``.
    """
    if not value:
        raise exception_type(f"{name or _DEFAULT_NAME} must be given")


def assert_instance(
    value: Any,
    dtype: type | tuple[type, ...],
    name: str | None = None,
    exception_type: type[Exception] = TypeError,
):
    """Assert *value* is an instance of data type *dtype*.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test.
        dtype: A type or tuple of types.
        name: Name of a variable that holds *value*.
        exception_type: The exception type. Default is ``TypeError``.
    """
    if not isinstance(value, dtype):
        raise exception_type(
            f"{name or _DEFAULT_NAME} must be an instance of {dtype}, was {type(value)}"
        )


def assert_in(
    value: Any,
    container: Container,
    name: str = None,
    exception_type: type[Exception] = ValueError,
):
    """Assert *value* is a member of *container*.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test for membership.
        container: The container.
        name: Name of a variable that holds *value*.
        exception_type: The exception type. Default is ``ValueError``.
    """
    if value not in container:
        raise exception_type(f"{name or _DEFAULT_NAME} must be one of {container}")


def assert_true(value: Any, message: str, exception_type: type[Exception] = ValueError):
    """Assert *value* is true after conversion into a Boolean value.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test.
        message: The error message used if the assertion fails.
        exception_type: The exception type. Default is ``ValueError``.
    """
    if not value:
        raise exception_type(message)



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from collections.abc import Sequence

import dask.array as da
import numpy as np
import pyproj
import xarray as xr

from xcube_resampling.constants import FloatInt

from .assertions import assert_true
from .base import GridMapping
from .helpers import (
    _default_xy_dim_names,
    _default_xy_var_names,
    _normalize_crs,
    _normalize_int_pair,
    _normalize_number_pair,
    _round_sequence,
    _to_int_or_float,
)


class RegularGridMapping(GridMapping):
    def __init__(self, **kwargs):
        kwargs.pop("is_regular", None)
        super().__init__(is_regular=True, **kwargs)
        self._xy_coords = None

    def _new_x_coords(self) -> xr.DataArray:
        self._assert_regular()
        return xr.DataArray(
            da.linspace(self.x_min, self.x_max, self.width, chunks=self.tile_width),
            dims=self.xy_dim_names[0],
            name=self.xy_var_names[0],
        )

    def _new_y_coords(self) -> xr.DataArray:
        self._assert_regular()
        if self.is_j_axis_up:
            return xr.DataArray(
                da.linspace(
                    self.y_min, self.y_max, self.height, chunks=self.tile_height
                ),
                dims=self.xy_dim_names[1],
                name=self.xy_var_names[1],
            )
        else:
            return xr.DataArray(
                da.linspace(
                    self.y_max, self.y_min, self.height, chunks=self.tile_height
                ),
                dims=self.xy_dim_names[1],
                name=self.xy_var_names[1],
            )

    def _new_xy_coords(self) -> xr.DataArray:
        self._assert_regular()
        x_coords_1d = da.asarray(self.x_coords.data).rechunk(self.tile_width)
        y_coords_1d = da.expand_dims(
            da.asarray(self.y_coords.data).rechunk(self.tile_height), 1
        )
        y_coords_2d, x_coords_2d = da.broadcast_arrays(y_coords_1d, x_coords_1d)
        xy_coords = da.concatenate(
            [da.expand_dims(x_coords_2d, 0), da.expand_dims(y_coords_2d, 0)]
        )
        xy_coords = da.rechunk(
            xy_coords, chunks=(2, xy_coords.chunksize[1], xy_coords.chunksize[2])
        )
        xy_coords = xr.DataArray(
            xy_coords,
            dims=("coord", self.y_coords.dims[0], self.x_coords.dims[0]),
            name="xy_coords",
        )
        xy_coords.name = "xy_coords"
        return xy_coords


def new_regular_grid_mapping(
    size: int | tuple[int, int],
    xy_min: tuple[float, float],
    xy_res: float | tuple[float, float],
    crs: str | pyproj.crs.CRS,
    *,
    tile_size: int | tuple[int, int] = None,
    is_j_axis_up: bool = False,
) -> GridMapping:
    width, height = _normalize_int_pair(size, name="size")
    assert_true(width > 1 and height > 1, "invalid size")

    x_min, y_min = _normalize_number_pair(xy_min, name="xy_min")

    x_res, y_res = _normalize_number_pair(xy_res, name="xy_res")
    assert_true(x_res > 0 and y_res > 0, "invalid xy_res")

    crs = _normalize_crs(crs)

    x_min = _to_int_or_float(x_min)
    y_min = _to_int_or_float(y_min)
    x_max = _to_int_or_float(x_min + x_res * (width - 1))
    y_max = _to_int_or_float(y_min + y_res * (height - 1))
    xy_bbox = x_min - x_res / 2, y_min - y_res / 2, x_max + x_res / 2, y_max + y_res / 2
    xy_bbox = _round_sequence(xy_bbox)
    if crs.is_geographic:
        if xy_bbox[1] < -90:
            raise ValueError("invalid xy_bbox (south)")
        if xy_bbox[3] > 90:
            raise ValueError("invalid xy_bbox (north)")

    return RegularGridMapping(
        crs=crs,
        size=(width, height),
        tile_size=tile_size or (width, height),
        xy_bbox=xy_bbox,
        xy_res=(x_res, y_res),
        xy_var_names=_default_xy_var_names(crs),
        xy_dim_names=_default_xy_dim_names(crs),
        is_lon_360=(x_max > 180) and crs.is_geographic,
        is_j_axis_up=is_j_axis_up,
    )


def to_regular_grid_mapping(
    grid_mapping: GridMapping,
    *,
    tile_size: int | tuple[int, int] = None,
    is_j_axis_up: bool = False,
) -> GridMapping:
    if grid_mapping.is_regular:
        if tile_size is not None or is_j_axis_up != grid_mapping.is_j_axis_up:
            return grid_mapping.derive(tile_size=tile_size, is_j_axis_up=is_j_axis_up)
        return grid_mapping

    x_res, y_res = grid_mapping.xy_res
    x_min = grid_mapping.xy_bbox[0] + x_res / 2
    y_min = grid_mapping.xy_bbox[1] + y_res / 2
    x_max = grid_mapping.xy_bbox[2] - x_res / 2
    y_max = grid_mapping.xy_bbox[3] - y_res / 2

    width = round((x_max - x_min + x_res) / x_res)
    height = round((y_max - y_min + y_res) / y_res)
    width = width if width >= 2 else 2
    height = height if height >= 2 else 2

    if tile_size is None:
        tile_size = grid_mapping.tile_size
    return new_regular_grid_mapping(
        size=(width, height),
        xy_min=(x_min, y_min),
        xy_res=(x_res, y_res),
        crs=grid_mapping.crs,
        tile_size=tile_size,
        is_j_axis_up=is_j_axis_up,
    )


def new_regular_grid_mapping_from_bbox(
    bbox: Sequence[FloatInt],
    xy_res: FloatInt | tuple[FloatInt, FloatInt],
    crs: str | pyproj.CRS,
    tile_size: int | tuple[int, int] = 1024,
    is_j_axis_up: bool = False,
) -> GridMapping:
    if not isinstance(xy_res, tuple):
        xy_res = (xy_res, xy_res)
    x_size = int(np.ceil((bbox[2] - bbox[0]) / xy_res[0]))
    y_size = int(np.ceil(abs(bbox[3] - bbox[1]) / xy_res[1]))
    return new_regular_grid_mapping(
        size=(x_size, y_size),
        xy_min=(bbox[0] + xy_res[0] / 2, bbox[1] + xy_res[1] / 2),
        xy_res=xy_res,
        crs=crs,
        tile_size=tile_size,
        is_j_axis_up=is_j_axis_up,
    )



import unittest

import dask.array as da
import numpy as np
import xarray as xr

from xcube_resampling.constants import (
    AGG_METHODS,
    FILLVALUE_INT,
    FILLVALUE_UINT8,
    FILLVALUE_UINT16,
    FILLVALUE_UINT32,
)
from xcube_resampling.gridmapping import GridMapping

# noinspection PyProtectedMember
from xcube_resampling.utils import (
    _create_empty_dataset,
    _get_fill_value,
    _get_grid_mapping_name,
    _get_prevent_nan_propagation,
    _get_spatial_agg_method,
    _get_spatial_interp_method,
    _prep_spatial_interp_methods_downscale,
    _select_variables,
    bbox_overlap,
    clip_dataset_by_bbox,
    get_spatial_coords,
    get_utm_crs,
    reproject_bbox,
    resolution_meters_to_degrees,
)

from .sampledata import create_2x4x4_dataset_with_irregular_coords


class TestUtils(unittest.TestCase):

    def test_get_spatial_coords_lon_lat(self):
        # Dataset with "lon" and "lat"
        ds = xr.Dataset(coords={"lon": [0, 1], "lat": [0, 1]})
        x_dim, y_dim = get_spatial_coords(ds)
        self.assertEqual((x_dim, y_dim), ("lon", "lat"))

    def test_get_spatial_coords_longitude_latitude(self):
        # Dataset with "longitude" and "latitude"
        ds = xr.Dataset(coords={"longitude": [0, 1], "latitude": [0, 1]})
        x_dim, y_dim = get_spatial_coords(ds)
        self.assertEqual((x_dim, y_dim), ("longitude", "latitude"))

    def test_get_spatial_coords_x_y(self):
        # Dataset with "x" and "y"
        ds = xr.Dataset(coords={"x": [0, 1], "y": [0, 1]})
        x_dim, y_dim = get_spatial_coords(ds)
        self.assertEqual((x_dim, y_dim), ("x", "y"))

    def test_get_spatial_coords_missing_dims(self):
        # Dataset with no recognized spatial dimensions
        ds = xr.Dataset(coords={"time": [0, 1]})
        with self.assertRaises(KeyError) as context:
            get_spatial_coords(ds)
        self.assertIn("No standard spatial coordinates found", str(context.exception))

    def test_get_utm_crs(self):
        crs = get_utm_crs(10, 55.5)
        self.assertEqual("EPSG:32632", crs.to_string())
        crs = get_utm_crs(5, 60)
        self.assertEqual("EPSG:32632", crs.to_string())
        crs = get_utm_crs(2.9, 60)
        self.assertEqual("EPSG:32631", crs.to_string())
        crs = get_utm_crs(8, 75)
        self.assertEqual("EPSG:32631", crs.to_string())
        crs = get_utm_crs(20, 75)
        self.assertEqual("EPSG:32633", crs.to_string())
        crs = get_utm_crs(22, 75)
        self.assertEqual("EPSG:32635", crs.to_string())
        crs = get_utm_crs(40, 75)
        self.assertEqual("EPSG:32637", crs.to_string())

    def test_select_variables(self):
        ds = xr.Dataset(
            {
                "var1": ("x", [1, 2, 3]),
                "var2": ("x", [4, 5, 6]),
                "var3": ("x", [7, 8, 9]),
            },
            coords={"x": [0, 1, 2]},
        )

        # if variables=None, return dataset with all data variables
        result = _select_variables(ds, variables=None)
        self.assertEqual(set(result.data_vars), set(ds.data_vars))

        # select one variable
        result = _select_variables(ds, variables="var1")
        self.assertEqual(list(result.data_vars), ["var1"])
        self.assertTrue("var1" in result)

        # select multiple variables
        result = _select_variables(ds, variables=["var1", "var3"])
        self.assertEqual(set(result.data_vars), {"var1", "var3"})
        self.assertTrue("var2" not in result)

        # selecting a variable not in dataset should raise KeyError
        with self.assertRaises(KeyError):
            _select_variables(ds, variables="nonexistent_var")

    def test_n_get_grid_mapping_name(self):
        # no grid mapping
        ds = xr.Dataset({"var1": ("x", [1, 2, 3])}, coords={"x": [0, 1, 2]})
        self.assertIsNone(_get_grid_mapping_name(ds))

        # grid mapping in variables attribute
        ds = xr.Dataset({"var1": ("x", [1, 2, 3])})
        ds["var1"].attrs["grid_mapping"] = "crs_var"
        self.assertEqual(_get_grid_mapping_name(ds), "crs_var")

        # grid mapping in crs variable
        ds = xr.Dataset({"var1": ("x", [1, 2, 3]), "crs": 0}, coords={"x": [0, 1, 2]})
        self.assertEqual(_get_grid_mapping_name(ds), "crs")

        # grid mapping in spatial ref coordinate
        ds = xr.Dataset(
            {"var1": ("x", [1, 2, 3])}, coords={"x": [0, 1, 2], "spatial_ref": 0}
        )
        self.assertEqual(_get_grid_mapping_name(ds), "spatial_ref")

        # if multiple grid mapping found, error should be raised.
        ds = xr.Dataset({"var1": ("x", [1, 2, 3])})
        ds["var1"].attrs["grid_mapping"] = "gm1"
        ds["crs"] = 0
        with self.assertRaises(AssertionError):
            _get_grid_mapping_name(ds)

    def test_get_interp_method(self):
        int_var = xr.DataArray(np.array([1, 2, 3], dtype=np.int32), dims=["x"])
        bool_var = xr.DataArray(
            np.array([True, False, True], dtype=np.bool_), dims=["x"]
        )
        float_var = xr.DataArray(
            np.array([1.0, 2.0, 3.0], dtype=np.float32), dims=["x"]
        )

        # integer type data array
        result = _get_spatial_interp_method(None, "var", int_var)
        self.assertEqual(result, 0)

        # bool type data array
        result = _get_spatial_interp_method(None, "var", bool_var)
        self.assertEqual(result, 0)

        # float type data array
        result = _get_spatial_interp_method(None, "var", float_var)
        self.assertEqual(result, 1)

        # integer scalar
        result = _get_spatial_interp_method(1, "var", float_var)
        self.assertEqual(result, 1)

        # string
        result = _get_spatial_interp_method("nearest", "var", int_var)
        self.assertEqual(result, "nearest")

        # key matching
        interp_methods = {"var": "bilinear"}
        # noinspection PyTypeChecker
        result = _get_spatial_interp_method(interp_methods, "var", float_var)
        self.assertEqual(result, "bilinear")

        # dtaa type matching
        interp_methods = {np.dtype("float32"): "bilinear"}
        # noinspection PyTypeChecker
        result = _get_spatial_interp_method(interp_methods, "other", float_var)
        self.assertEqual(result, "bilinear")

        # no matching keys shall trigger a log warning
        interp_methods = {"something": "bilinear"}
        with self.assertLogs("xcube.resampling", level="DEBUG") as cm:
            # noinspection PyTypeChecker
            result = _get_spatial_interp_method(interp_methods, "var", int_var)
        self.assertEqual(result, 0)  # default value
        self.assertIn("Defaults are assigned", cm.output[0])

    def test_prep_interp_methods_downscale(self):
        self.assertIsNone(_prep_spatial_interp_methods_downscale(None))
        self.assertEqual(
            _prep_spatial_interp_methods_downscale("triangular"), "bilinear"
        )
        self.assertEqual(_prep_spatial_interp_methods_downscale("nearest"), "nearest")
        self.assertEqual(_prep_spatial_interp_methods_downscale(1), 1)

        interp_map = {"a": "triangular", "b": "nearest"}
        expected = {"a": "bilinear", "b": "nearest"}
        # noinspection PyTypeChecker
        self.assertEqual(
            _prep_spatial_interp_methods_downscale(interp_map),
            expected,
        )

        interp_map = {"a": "nearest", "b": "bilinear"}
        # noinspection PyTypeChecker
        self.assertEqual(
            _prep_spatial_interp_methods_downscale(interp_map),
            interp_map,
        )

    def test_get_agg_method(self):
        int_var = xr.DataArray(np.array([1, 2, 3], dtype=np.int32), dims=["x"])
        bool_var = xr.DataArray(
            np.array([True, False, True], dtype=np.bool_), dims=["x"]
        )
        float_var = xr.DataArray(
            np.array([1.0, 2.0, 3.0], dtype=np.float32), dims=["x"]
        )

        # integer type data array, default
        result = _get_spatial_agg_method(None, "var", int_var)
        self.assertEqual(result, AGG_METHODS["center"])

        # bool type data array, default
        result = _get_spatial_agg_method(None, "var", bool_var)
        self.assertEqual(result, AGG_METHODS["center"])

        # float type data array, default
        result = _get_spatial_agg_method(None, "var", float_var)
        self.assertEqual(result, AGG_METHODS["mean"])

        # string as method
        result = _get_spatial_agg_method("center", "var", float_var)
        self.assertEqual(result, AGG_METHODS["center"])

        # key matching
        agg_methods = {"var": "mean"}
        # noinspection PyTypeChecker
        result = _get_spatial_agg_method(agg_methods, "var", int_var)
        self.assertEqual(result, AGG_METHODS["mean"])

        # data type matching
        agg_methods = {np.dtype("float32"): "mean"}
        # noinspection PyTypeChecker
        result = _get_spatial_agg_method(agg_methods, "other", float_var)
        self.assertEqual(result, AGG_METHODS["mean"])

        # no matching keys triggers log warning
        agg_methods = {"something": "mean"}
        with self.assertLogs("xcube.resampling", level="DEBUG") as cm:
            # noinspection PyTypeChecker
            result = _get_spatial_agg_method(agg_methods, "var", int_var)
        self.assertEqual(result, AGG_METHODS["center"])  # default value
        self.assertIn("Defaults are assigned", cm.output[0])

    def test_get_prevent_nan_propagation(self):
        int_var = xr.DataArray(np.array([1, 2, 3], dtype=np.int32), dims=["x"])
        float_var = xr.DataArray(
            np.array([1.0, 2.0, 3.0], dtype=np.float32), dims=["x"]
        )

        # bool directly
        result = _get_prevent_nan_propagation(True, "var", int_var)
        self.assertTrue(result)

        result = _get_prevent_nan_propagation(False, "var", float_var)
        self.assertFalse(result)

        # key mapping
        prevent_nan_propagations = {"var": True}
        # noinspection PyTypeChecker
        result = _get_prevent_nan_propagation(prevent_nan_propagations, "var", int_var)
        self.assertTrue(result)

        # dtype mapping
        prevent_nan_propagations = {np.dtype("float32"): True}
        # noinspection PyTypeChecker
        result = _get_prevent_nan_propagation(
            prevent_nan_propagations, "other", float_var
        )
        self.assertTrue(result)

        # missing key/dtype → default False with log warning
        prevent_nan_propagations = {"something": True}
        with self.assertLogs("xcube.resampling", level="DEBUG") as cm:
            # noinspection PyTypeChecker
            result = _get_prevent_nan_propagation(
                prevent_nan_propagations, "var", int_var
            )
        self.assertFalse(result)
        self.assertIn("Defaults are assigned", cm.output[0])

        # prevent_nan_propagations is None → fallback default False
        result = _get_prevent_nan_propagation(None, "var", float_var)
        self.assertFalse(result)

    def test_get_fill_value(self):
        uint8_var = xr.DataArray(np.array([1, 2, 3], dtype=np.uint8), dims=["x"])
        uint16_var = xr.DataArray(np.array([1, 2, 3], dtype=np.uint16), dims=["x"])
        uint32_var = xr.DataArray(np.array([1, 2, 3], dtype=np.uint32), dims=["x"])
        int_var = xr.DataArray(np.array([1, 2, 3], dtype=np.int32), dims=["x"])
        bool_var = xr.DataArray(
            np.array([True, False, True], dtype=np.bool_), dims=["x"]
        )
        float_var = xr.DataArray(
            np.array([1.0, 2.0, 3.0], dtype=np.float32), dims=["x"]
        )

        # scalar int
        result = _get_fill_value(-99, "var", int_var)
        self.assertEqual(result, -99)

        # scalar float
        result = _get_fill_value(-9.9, "var", float_var)
        self.assertEqual(result, -9.9)

        # mapping by variable name
        result = _get_fill_value({"var": 1234}, "var", int_var)
        self.assertEqual(result, 1234)

        # mapping by dtype
        result = _get_fill_value({np.dtype("float32"): 3.14}, "other", float_var)
        self.assertEqual(result, 3.14)

        # unmatched mapping triggers warning + defaults
        with self.assertLogs("xcube.resampling", level="DEBUG") as cm:
            result = _get_fill_value({"something": 42}, "var", int_var)
        self.assertEqual(result, FILLVALUE_INT)
        self.assertIn("Fill value could not be derived", cm.output[0])

        # defaults
        self.assertEqual(_get_fill_value(None, "var", uint8_var), FILLVALUE_UINT8)
        self.assertEqual(_get_fill_value(None, "var", uint16_var), FILLVALUE_UINT16)
        self.assertEqual(_get_fill_value(None, "var", uint32_var), FILLVALUE_UINT32)
        self.assertEqual(_get_fill_value(None, "var", int_var), FILLVALUE_INT)
        self.assertEqual(_get_fill_value(None, "var", bool_var), 0)
        self.assertTrue(np.isnan(_get_fill_value(None, "var", float_var)))

    def test_reproject_bbox(self):
        bbox_wgs84 = [2, 50, 3, 51]
        crs_wgs84 = "EPSG:4326"
        crs_3035 = "EPSG:3035"
        bbox_3035 = [3748675.9529771, 3011432.8944597, 3830472.1359979, 3129432.4914285]
        self.assertEqual(bbox_wgs84, reproject_bbox(bbox_wgs84, crs_wgs84, crs_wgs84))
        self.assertEqual(bbox_3035, reproject_bbox(bbox_3035, crs_3035, crs_3035))
        np.testing.assert_almost_equal(
            reproject_bbox(bbox_wgs84, crs_wgs84, crs_3035), bbox_3035
        )
        np.testing.assert_almost_equal(
            reproject_bbox(
                reproject_bbox(bbox_wgs84, crs_wgs84, crs_3035), crs_3035, crs_wgs84
            ),
            [
                1.829619451017442,
                49.93464594063249,
                3.1462425554926226,
                51.06428203128216,
            ],
        )

    def test_bbox_overlap(self):
        # identical boxes
        bbox = (0, 0, 10, 10)
        self.assertEqual(1.0, bbox_overlap(bbox, bbox))

        # partial overlap
        source = (0, 0, 10, 10)
        target = (5, 5, 15, 15)
        # overlap area = 25, source area = 100
        expected = 25 / 100
        self.assertAlmostEqual(expected, bbox_overlap(source, target))

        # no overlap
        source = (0, 0, 10, 10)
        target = (20, 20, 30, 30)
        self.assertAlmostEqual(0.0, bbox_overlap(source, target))

        # target fully inside source
        source = (0, 0, 10, 10)
        target = (2, 2, 8, 8)
        # overlap area = 36, source area = 100
        expected = 0.36
        self.assertAlmostEqual(expected, bbox_overlap(source, target))

        # source fully inside target
        source = (2, 2, 8, 8)
        target = (0, 0, 10, 10)
        # overlap area = source area = 36
        expected = 36 / 36
        self.assertAlmostEqual(expected, bbox_overlap(source, target))

        # antimeridian - identical wrapped boxes (e.g. 170°E → -170°W)
        source = (170, -10, -170, 10)
        target = (170, -10, -170, 10)
        self.assertEqual(1.0, bbox_overlap(source, target))

        # antimeridian - source crosses, target fully inside one side
        source = (170, 0, -170, 10)
        target = (175, 0, 178, 10)
        expected = 30 / 200
        self.assertAlmostEqual(expected, bbox_overlap(source, target))

        # antimeridian - partial overlap across both segments
        source = (170, 0, -170, 10)
        target = (160, 0, 175, 10)
        expected = 50 / 200
        self.assertAlmostEqual(expected, bbox_overlap(source, target))

        # antimeridian - non-wrapped target covering entire world
        source = (170, 0, -170, 10)
        target = (-180, -90, 180, 90)
        self.assertEqual(1.0, bbox_overlap(source, target))

        # antimeridian - wrapped target covering entire world
        source = (170, 0, -170, 10)
        target = (170, -90, 169.99, 90)
        self.assertEqual(1.0, bbox_overlap(source, target))

        # antimeridian - touching at boundary only (no real overlap)
        source = (170, 0, -170, 10)
        target = (-170, 0, -160, 10)
        self.assertAlmostEqual(0.0, bbox_overlap(source, target))

    def test_resolution_meters_to_degrees(self):
        # 111320 m ≈ 1 degree at equator
        lat_deg, lon_deg = resolution_meters_to_degrees(111320, 0)
        self.assertAlmostEqual(1.0, lat_deg, places=6)
        self.assertAlmostEqual(1.0, lon_deg, places=6)

        # 222640 m ≈ 2 degrees latitude
        lon_deg, lat_deg = resolution_meters_to_degrees((111320, 222640), 0)
        self.assertAlmostEqual(2.0, lat_deg, places=6)
        self.assertAlmostEqual(1.0, lon_deg, places=6)

        # At 60 degrees latitude, longitude degrees shrink by cos(60°) = 0.5
        lon_deg, lat_deg = resolution_meters_to_degrees(111320, 60)
        self.assertAlmostEqual(1.0, lat_deg, places=6)
        self.assertAlmostEqual(1.0 / 0.5, lon_deg, places=6)  # 2 degrees


class TestClipDatasetByBBox(unittest.TestCase):

    def setUp(self):
        # 1D coordinates dataset
        x = np.linspace(1, 10, 10)
        y = np.linspace(1, 20, 20)
        data_1d = np.random.rand(len(y), len(x))
        self.ds_1d = xr.Dataset({"var": (("y", "x"), data_1d)}, coords={"x": x, "y": y})

        # 2D coordinates dataset
        x2d, y2d = np.meshgrid(x, y)
        data_2d = np.random.rand(*x2d.shape)
        self.ds_2d = xr.Dataset(
            {"var": (("y", "x"), data_2d)},
            coords={"lon": (("y", "x"), x2d), "lat": (("y", "x"), y2d)},
        )

        # 2D coordinates dataset as dark array
        x2d, y2d = np.meshgrid(x, y)
        data_2d = np.random.rand(*x2d.shape)
        self.ds_2d_chunked = xr.Dataset(
            {"var": (("row", "column"), data_2d)},
            coords={
                "lon": (("row", "column"), da.from_array(x2d, chunks=(5, 5))),
                "lat": (("row", "column"), da.from_array(y2d, chunks=(5, 5))),
            },
        )

    def test_clip_1dcoord_inside_bbox(self):
        bbox = [2, 5, 8, 15]  # xmin, ymin, xmax, ymax
        clipped = clip_dataset_by_bbox(self.ds_1d, bbox, spatial_coords=("x", "y"))
        self.assertTrue((clipped.x >= bbox[0]).all())
        self.assertTrue((clipped.x <= bbox[2]).all())
        self.assertTrue((clipped.y >= bbox[1]).all())
        self.assertTrue((clipped.y <= bbox[3]).all())

    def test_clip_2dcoord_inside_bbox(self):
        bbox = [2, 5, 8, 15]
        clipped = clip_dataset_by_bbox(self.ds_2d, bbox)
        self.assertTrue((clipped["lon"].values >= bbox[0]).all())
        self.assertTrue((clipped["lon"].values <= bbox[2]).all())
        self.assertTrue((clipped["lat"].values >= bbox[1]).all())
        self.assertTrue((clipped["lat"].values <= bbox[3]).all())

    def test_clip_2dcoord_inside_bbox_chunked(self):
        bbox = [2, 5, 8, 15]
        clipped = clip_dataset_by_bbox(self.ds_2d_chunked, bbox)
        self.assertTrue((clipped["lon"].values >= bbox[0]).all())
        self.assertTrue((clipped["lon"].values <= bbox[2]).all())
        self.assertTrue((clipped["lat"].values >= bbox[1]).all())
        self.assertTrue((clipped["lat"].values <= bbox[3]).all())

    def test_clip_dataset_by_bbox_invalid_bbox(self):
        with self.assertRaises(ValueError) as context:
            clip_dataset_by_bbox(self.ds_1d, bbox=[0, 0, 1])
        self.assertIn("Expected bbox of length 4", str(context.exception))

    def test_unsupported_coord_dims(self):
        ds = self.ds_1d.copy()
        ds["x"] = ds["x"].expand_dims("z")  # 2D+ coordinate
        with self.assertRaises(ValueError):
            clip_dataset_by_bbox(ds, [0, 0, 5, 5])

    def test_bbox_outside_1d_dataset(self):
        bbox = [100, 100, 110, 110]  # completely outside
        with self.assertLogs("xcube.resampling", level="WARNING") as cm:
            clipped = clip_dataset_by_bbox(self.ds_1d, bbox)
        self.assertIn(
            "Clipped dataset contains at least one zero-sized dimension.", cm.output[0]
        )
        # should result in zero-sized dimensions
        self.assertTrue(any(size == 0 for size in clipped.sizes.values()))

    def test_bbox_outside_2d_dataset(self):
        bbox = [100, 100, 110, 110]  # completely outside
        with self.assertLogs("xcube.resampling", level="WARNING") as cm:
            clipped = clip_dataset_by_bbox(self.ds_2d, bbox)
        self.assertIn(
            "Clipped dataset contains at least one zero-sized dimension.", cm.output[0]
        )
        # should result in zero-sized dimensions
        self.assertTrue(any(size == 0 for size in clipped.sizes.values()))

    def test_create_empty_dataset_3d(self):
        source_ds = create_2x4x4_dataset_with_irregular_coords()
        source_ds = source_ds.chunk(dict(y=2, x=2))
        source_gm = GridMapping.from_dataset(source_ds)
        target_gm = GridMapping.regular(
            size=(3, 3), xy_min=(0.0, 0.0), xy_res=0.1, crs="epsg:4326"
        )
        target_ds = _create_empty_dataset(source_ds, source_gm, target_gm)
        self.assertCountEqual(["rad"], target_ds.data_vars)
        self.assertCountEqual(("time", "lat", "lon"), target_ds["rad"].dims)
        self.assertCountEqual((2, 3, 3), target_ds["rad"].shape)
        np.testing.assert_array_equal(
            np.isnan(target_ds.rad), np.ones_like(target_ds.rad, dtype=bool)
        )



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest

import numpy as np
import pyproj
import xarray as xr

from xcube_resampling.affine import affine_transform_dataset
from xcube_resampling.gridmapping import CRS_CRS84, CRS_WGS84, GridMapping

from .sampledata import (
    create_2x8x6_dataset_with_regular_coords,
    create_8x6_dataset_with_regular_coords,
)


class AffineTransformDatasetTest(unittest.TestCase):
    def setUp(self):
        self.source_ds = create_8x6_dataset_with_regular_coords()
        self.source_ds_3d = create_2x8x6_dataset_with_regular_coords()

        self.source_gm = GridMapping.from_dataset(self.source_ds)
        self.res = 0.1

    def test_subset(self):
        target_gm = GridMapping.regular(
            (3, 3), (50.05, 10.05), self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            interp_methods=1,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [1, 0, 2],
                    [0, 3, 0],
                    [4, 0, 1],
                ]
            ),
        )

        target_gm = GridMapping.regular(
            (3, 3), (50.15, 10.15), self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds, target_gm, interp_methods=1
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [4, np.nan, np.nan],
                    [0, 2, 0],
                    [3, 0, 4],
                ]
            ),
        )

        target_gm = GridMapping.regular(
            (3, 3), (50.1, 10.1), self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds, target_gm, interp_methods=1
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [1.25, 1.5, np.nan],
                    [1.0, 1.25, 1.5],
                    [1.75, 1.0, 1.25],
                ]
            ),
        )

        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods=1,
            prevent_nan_propagations=True,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [1.25, 1.5, 0.6666667],
                    [1.0, 1.25, 1.5],
                    [1.75, 1.0, 1.25],
                ]
            ),
        )

    def test_subset_3d(self):
        target_gm = GridMapping.regular(
            (3, 3), (50.05, 10.05), self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds_3d,
            target_gm,
            interp_methods=1,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds_3d.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((2, 3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [
                        [1, 0, 2],
                        [0, 3, 0],
                        [4, 0, 1],
                    ],
                    [
                        [1, 0, 2],
                        [0, 3, 0],
                        [4, 0, 1],
                    ],
                ]
            ),
        )

    def test_subset_with_source_gm(self):
        target_gm = GridMapping.regular(
            (3, 3), (50.05, 10.05), self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods="bilinear",
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [1, 0, 2],
                    [0, 3, 0],
                    [4, 0, 1],
                ]
            ),
        )

        target_gm = GridMapping.regular(
            (3, 3), (50.15, 10.15), self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods={"refl": "bilinear"},
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [4, np.nan, np.nan],
                    [0, 2, 0],
                    [3, 0, 4],
                ]
            ),
        )

        target_gm = GridMapping.regular(
            (3, 3), (50.1, 10.1), self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods={"refl": 1},
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array([[1.25, 1.5, np.nan], [1.0, 1.25, 1.5], [1.75, 1.0, 1.25]]),
        )

    def test_different_geographic_crses(self):
        expected = np.array([[1.0, 0.0, 2.0], [0.0, 3.0, 0.0], [4.0, 0.0, 1.0]])

        target_gm = GridMapping.regular((3, 3), (50.05, 10.05), self.res, CRS_WGS84)
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods=1,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(target_ds.refl.values, expected)

        target_gm = GridMapping.regular((3, 3), (50.05, 10.05), self.res, CRS_CRS84)
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods=1,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(target_ds.refl.values, expected)

        target_gm = GridMapping.regular(
            (3, 3), (50.05, 10.05), self.res, pyproj.crs.CRS(3035)
        )
        with self.assertRaises(AssertionError) as cm:
            affine_transform_dataset(
                self.source_ds, target_gm, source_gm=self.source_gm
            )

        self.assertIn(
            "Affine transformation cannot be applied to source CRS 'WGS 84' "
            "and target CRS 'ETRS89-extended / LAEA Europe'",
            str(cm.exception),
        )

    def test_downscale_x2(self):
        target_gm = GridMapping.regular(
            (8, 6), (50.05, 10.15), 2 * self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods=1,
            prevent_nan_propagations=True,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((6, 8), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                    [0.75, 1.25, 1.75, 1.25, np.nan, np.nan, np.nan, np.nan],
                    [1.25, 2 / 3, 1.25, 1.75, np.nan, np.nan, np.nan, np.nan],
                    [1.75, 1.25, 0.75, 1.25, np.nan, np.nan, np.nan, np.nan],
                ]
            ),
        )

    def test_downscale_x2_and_shift(self):
        target_gm = GridMapping.regular(
            (8, 6), (49.85, 9.95), 2 * self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods=1,
            prevent_nan_propagations=True,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((6, 8), target_ds.refl.shape)
        print(repr(target_ds.refl.values))
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, 0.75, 1.25, 1.75, 1.25, np.nan, np.nan, np.nan],
                    [np.nan, 1.25, 2 / 3, 1.25, 1.75, np.nan, np.nan, np.nan],
                    [np.nan, 1.75, 1.25, 0.75, 1.25, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                ]
            ),
        )

    def test_upscale_x2(self):
        target_gm = GridMapping.regular(
            (8, 6), (50.05, 10.0), self.res / 2, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods=1,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((6, 8), target_ds.refl.shape)
        print(repr(target_ds.refl.values))
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [1.0, 0.5, 0.0, 1.0, 2.0, 1.0, 0.0, 1.5],
                    [0.5, 1.0, 1.5, 1.25, 1.0, 1.5, 2.0, 1.75],
                    [0.0, 1.5, 3.0, 1.5, 0.0, 2.0, 4.0, 2.0],
                    [2.0, 1.75, 1.5, 1.0, 0.5, 1.25, 2.0, 1.5],
                    [4.0, 2.0, 0.0, 0.5, 1.0, 0.5, 0.0, 1.0],
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                ]
            ),
        )

    def test_upscale_x2_and_shift(self):
        target_gm = GridMapping.regular(
            (8, 6), (49.95, 9.95), self.res / 2, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods=1,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((6, 8), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [np.nan, np.nan, 0.5, 1.0, 1.5, 1.25, 1.0, 1.5],
                    [np.nan, np.nan, 0.0, 1.5, 3.0, 1.5, 0.0, 2.0],
                    [np.nan, np.nan, 2.0, 1.75, 1.5, 1.0, 0.5, 1.25],
                    [np.nan, np.nan, 4.0, 2.0, 0.0, 0.5, 1.0, 0.5],
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                ]
            ),
        )

    def test_shift(self):
        target_gm = GridMapping.regular(
            (8, 6), (50.25, 10.15), self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods=1,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((6, 8), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                    [0.0, 2.0, 0.0, 3.0, 0.0, 4.0, np.nan, np.nan],
                    [np.nan, np.nan, 4.0, 0.0, 1.0, 0.0, np.nan, np.nan],
                    [np.nan, np.nan, 0.0, 2.0, 0.0, 3.0, np.nan, np.nan],
                    [2.0, 0.0, 3.0, 0.0, 4.0, 0.0, np.nan, np.nan],
                    [0.0, 4.0, 0.0, 1.0, 0.0, 2.0, np.nan, np.nan],
                ]
            ),
        )

        target_gm = GridMapping.regular(
            (8, 6), (49.85, 9.95), self.res, self.source_gm.crs
        )
        target_ds = affine_transform_dataset(
            self.source_ds,
            target_gm,
            source_gm=self.source_gm,
            interp_methods=1,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(self.source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((6, 8), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [np.nan, np.nan, 2.0, 0.0, np.nan, np.nan, 4.0, 0.0],
                    [np.nan, np.nan, 0.0, 4.0, np.nan, np.nan, 0.0, 2.0],
                    [np.nan, np.nan, 1.0, 0.0, 2.0, 0.0, 3.0, 0.0],
                    [np.nan, np.nan, 0.0, 3.0, 0.0, 4.0, 0.0, 1.0],
                    [np.nan, np.nan, 4.0, 0.0, 1.0, 0.0, 2.0, 0.0],
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                ]
            ),
        )

    def test_affine_raise_value_error(self):
        target_gm = GridMapping.regular(
            (8, 6), (50.2, 10.1), self.res, self.source_gm.crs
        )
        with self.assertRaises(ValueError) as cm:
            _ = affine_transform_dataset(
                self.source_ds,
                target_gm,
                source_gm=self.source_gm,
                interp_methods=3,
            )
        self.assertIn(
            "interp_methods must be one of 0, 1, 'nearest', 'bilinear'. "
            "Higher order is not supported for 3D arrays in affine transforms, "
            "as it causes unintended blending across the non-spatial (e.g., time) "
            "dimension.",
            str(cm.exception),
        )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import unittest

import numpy as np
import xarray as xr

from xcube_resampling.gridmapping import CRS_WGS84, GridMapping
from xcube_resampling.spatial import resample_in_space

from .sampledata import (
    create_2x2_dataset_with_irregular_coords,
    create_4x4_dataset_with_irregular_coords,
    create_5x5_dataset_regular_utm,
    create_8x6_dataset_with_regular_coords,
)

nan = np.nan


# noinspection PyMethodMayBeStatic
class ResampleInSpaceTest(unittest.TestCase):
    def test_affine_transform_dataset(self):
        source_ds = create_8x6_dataset_with_regular_coords()
        source_gm = GridMapping.from_dataset(source_ds)
        target_gm = GridMapping.regular((3, 3), (50.05, 10.05), 0.1, source_gm.crs)
        target_ds = resample_in_space(
            source_ds,
            target_gm=target_gm,
            interp_methods=1,
        )
        self.assertIsInstance(target_ds, xr.Dataset)
        self.assertEqual(
            set(source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        self.assertEqual((3, 3), target_ds.refl.shape)
        np.testing.assert_almost_equal(
            target_ds.refl.values,
            np.array(
                [
                    [1, 0, 2],
                    [0, 3, 0],
                    [4, 0, 1],
                ]
            ),
        )

    def test_rectify_and_downscale_dataset(self):
        source_ds = create_4x4_dataset_with_irregular_coords()

        # nearest neighbour interpolation
        target_gm = GridMapping.regular(
            size=(2, 2), xy_min=(0, 52), xy_res=2, crs=CRS_WGS84
        )
        target_ds = resample_in_space(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_almost_equal(
            target_ds.rad.values,
            np.array(
                [
                    [5, 2],
                    [10, 7],
                ],
                dtype=target_ds.rad.dtype,
            ),
        )
        # bi-linear interpolation
        target_ds = resample_in_space(source_ds, target_gm=target_gm, interp_methods=1)
        np.testing.assert_almost_equal(
            target_ds.rad.values,
            np.array(
                [
                    [7.5, 4.5],
                    [12.5, 9.5],
                ],
                dtype=target_ds.rad.dtype,
            ),
        )

        # shifted target grid
        target_gm = GridMapping.regular(
            size=(2, 2), xy_min=(0, 50), xy_res=2, crs=CRS_WGS84
        )
        target_ds = resample_in_space(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_almost_equal(
            target_ds.rad.values,
            np.array(
                [
                    [10, 7],
                    [nan, nan],
                ],
                dtype=target_ds.rad.dtype,
            ),
        )

    def test_rectify_and_upscale_dataset(self):
        source_ds = create_2x2_dataset_with_irregular_coords()
        target_gm = GridMapping.regular(
            size=(4, 4), xy_min=(0, 50), xy_res=2, crs=CRS_WGS84
        )
        target_ds = resample_in_space(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_almost_equal(
            target_ds.rad.values,
            np.array(
                [
                    [nan, nan, nan, nan],
                    [nan, 1.0, 2.0, nan],
                    [3.0, 3.0, 2.0, nan],
                    [nan, 4.0, nan, nan],
                ],
                dtype=target_ds.rad.dtype,
            ),
        )

    def test_reproject_dataset(self):
        source_ds = create_5x5_dataset_regular_utm()

        # test projected CRS similar resolution
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(4320120, 3382520), xy_res=80, crs="epsg:3035"
        )
        target_ds = resample_in_space(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [1, 1, 2, 3, 4],
                    [6, 6, 7, 8, 9],
                    [11, 12, 12, 13, 14],
                    [16, 17, 17, 18, 19],
                    [21, 17, 17, 18, 19],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

        # test projected CRS finer resolution
        # test if subset calculation works as expected
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(4320090, 3382490), xy_res=20, crs="epsg:3035"
        )
        target_ds = resample_in_space(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [15, 16, 16, 16, 16],
                    [15, 16, 16, 16, 16],
                    [15, 16, 16, 16, 16],
                    [20, 21, 21, 21, 21],
                    [20, 21, 21, 21, 21],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

        # test geographic CRS with similar resolution
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(9.9889, 53.5502), xy_res=0.0006, crs=CRS_WGS84
        )
        target_ds = resample_in_space(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [7, 8, 8, 8, 9],
                    [12, 13, 13, 13, 14],
                    [12, 13, 13, 13, 14],
                    [17, 18, 18, 18, 19],
                    [22, 23, 23, 23, 24],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

        # test geographic CRS with 1/2 resolution
        # test if subset calculation works as expected
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(9.98875, 53.55005), xy_res=0.0003, crs=CRS_WGS84
        )
        target_ds = resample_in_space(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [12, 12, 12, 13, 13],
                    [17, 17, 17, 18, 18],
                    [17, 17, 17, 18, 18],
                    [22, 17, 17, 18, 18],
                    [22, 22, 22, 23, 23],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

    def test_resample_in_space_raise_logs(self):
        source_ds = create_5x5_dataset_regular_utm()
        with self.assertLogs("xcube.resampling", level="WARNING") as cm:
            _ = resample_in_space(source_ds)
        self.assertIn(
            "If source grid mapping is regular `target_gm` must be given. "
            "Source dataset is returned.",
            cm.output[0],
        )

    def test_resample_in_space_return_input_dataset(self):
        source_ds = create_5x5_dataset_regular_utm()
        target_gm = GridMapping.from_dataset(source_ds)
        target_ds = resample_in_space(source_ds, target_gm=target_gm)
        xr.testing.assert_equal(target_ds, source_ds)



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest

import numpy as np
import pyproj
import xarray as xr

from xcube_resampling.gridmapping import CRS_WGS84, GridMapping
from xcube_resampling.gridmapping.regular import RegularGridMapping

GEO_CRS = pyproj.crs.CRS(4326)
NOT_A_GEO_CRS = pyproj.crs.CRS(5243)


# noinspection PyMethodMayBeStatic
class RegularGridMappingTest(unittest.TestCase):
    def test_default_props(self):
        gm = GridMapping.regular((1000, 1000), (10, 53), 0.01, CRS_WGS84)
        self.assertEqual((1000, 1000), gm.size)
        self.assertEqual((1000, 1000), gm.tile_size)
        self.assertEqual(10, gm.x_min)
        self.assertEqual(53, gm.y_min)
        self.assertEqual((0.01, 0.01), gm.xy_res)
        self.assertEqual(True, gm.is_regular)
        self.assertEqual(False, gm.is_j_axis_up)

    def test_invalid_y(self):
        with self.assertRaises(ValueError) as cm:
            GridMapping.regular((1000, 1000), (10, -90.5), 0.01, CRS_WGS84)
        self.assertEqual("invalid xy_bbox (south)", f"{cm.exception}")

        with self.assertRaises(ValueError) as cm:
            GridMapping.regular((1000, 1000), (10, 53), 0.1, CRS_WGS84)
        self.assertEqual("invalid xy_bbox (north)", f"{cm.exception}")

    def test_xy_bbox(self):
        gm = GridMapping.regular((1000, 1000), (10, 53), 0.01, CRS_WGS84)
        np.testing.assert_allclose(gm.xy_bbox, (9.995, 52.995, 19.995, 62.995))
        self.assertEqual(False, gm.is_lon_360)

    def test_xy_bbox_anti_meridian(self):
        gm = GridMapping.regular((2000, 1000), (174.0, -30.0), 0.005, CRS_WGS84)
        np.testing.assert_allclose(gm.xy_bbox, (173.9975, -30.0025, 183.9975, -25.0025))
        self.assertEqual(True, gm.is_lon_360)

    def test_derive(self):
        gm = GridMapping.regular((1000, 1000), (10, 53), 0.01, CRS_WGS84)
        self.assertEqual((1000, 1000), gm.size)
        self.assertEqual((1000, 1000), gm.tile_size)
        self.assertEqual(False, gm.is_j_axis_up)
        derived_gm = gm.derive(tile_size=500, is_j_axis_up=True)
        self.assertIsNot(gm, derived_gm)
        self.assertIsInstance(derived_gm, RegularGridMapping)
        self.assertEqual((1000, 1000), derived_gm.size)
        self.assertEqual((500, 500), derived_gm.tile_size)
        self.assertEqual(True, derived_gm.is_j_axis_up)

    def test_xy_coords(self):
        gm = GridMapping.regular((8, 4), (10, 53), 0.1, CRS_WGS84).derive(
            tile_size=(4, 2)
        )
        xy_coords = gm.xy_coords
        self.assertIsInstance(xy_coords, xr.DataArray)
        self.assertIs(gm.xy_coords, xy_coords)
        self.assertEqual(("coord", "lat", "lon"), xy_coords.dims)
        self.assertEqual((2, 4, 8), xy_coords.shape)
        self.assertEqual(((2,), (2, 2), (4, 4)), xy_coords.chunks)
        np.testing.assert_almost_equal(
            xy_coords.values[0],
            np.array(
                [
                    [10.0, 10.1, 10.2, 10.3, 10.4, 10.5, 10.6, 10.7],
                    [10.0, 10.1, 10.2, 10.3, 10.4, 10.5, 10.6, 10.7],
                    [10.0, 10.1, 10.2, 10.3, 10.4, 10.5, 10.6, 10.7],
                    [10.0, 10.1, 10.2, 10.3, 10.4, 10.5, 10.6, 10.7],
                ]
            ),
        )
        np.testing.assert_almost_equal(
            xy_coords.values[1],
            np.array(
                [
                    [53.3, 53.3, 53.3, 53.3, 53.3, 53.3, 53.3, 53.3],
                    [53.2, 53.2, 53.2, 53.2, 53.2, 53.2, 53.2, 53.2],
                    [53.1, 53.1, 53.1, 53.1, 53.1, 53.1, 53.1, 53.1],
                    [53.0, 53.0, 53.0, 53.0, 53.0, 53.0, 53.0, 53.0],
                ]
            ),
        )

    def test_xy_names(self):
        gm = GridMapping.regular((1000, 1000), (10, 53), 0.01, GEO_CRS).derive(
            tile_size=500
        )
        self.assertEqual(("lon", "lat"), gm.xy_var_names)
        self.assertEqual(("lon", "lat"), gm.xy_dim_names)
        gm = GridMapping.regular((1000, 1000), (10, 53), 0.01, NOT_A_GEO_CRS).derive(
            tile_size=500
        )
        self.assertEqual(("x", "y"), gm.xy_var_names)
        self.assertEqual(("x", "y"), gm.xy_dim_names)

    def test_ij_bboxes(self):
        gm = GridMapping.regular(
            size=(2000, 1000), xy_min=(10.0, 20.0), xy_res=0.1, crs=NOT_A_GEO_CRS
        )
        np.testing.assert_almost_equal(
            gm.ij_bboxes, np.array([[0, 0, 2000, 1000]], dtype=np.int64)
        )

        gm = GridMapping.regular(
            size=(2000, 1000), xy_min=(10.0, 20.0), xy_res=0.1, crs=NOT_A_GEO_CRS
        ).derive(tile_size=500)
        np.testing.assert_almost_equal(
            gm.ij_bboxes,
            np.array(
                [
                    [0, 0, 500, 500],
                    [500, 0, 1000, 500],
                    [1000, 0, 1500, 500],
                    [1500, 0, 2000, 500],
                    [0, 500, 500, 1000],
                    [500, 500, 1000, 1000],
                    [1000, 500, 1500, 1000],
                    [1500, 500, 2000, 1000],
                ],
                dtype=np.int64,
            ),
        )

    def test_xy_bboxes(self):
        gm = GridMapping.regular(
            size=(2000, 1000), xy_min=(10.0, 20.0), xy_res=0.1, crs=NOT_A_GEO_CRS
        )
        np.testing.assert_almost_equal(
            gm.xy_bboxes, np.array([[9.95, 19.95, 209.95, 119.95]], dtype=np.float64)
        )

        gm = GridMapping.regular(
            size=(2000, 1000), xy_min=(10.0, 20.0), xy_res=0.1, crs=NOT_A_GEO_CRS
        ).derive(tile_size=500)
        np.testing.assert_almost_equal(
            gm.xy_bboxes,
            np.array(
                [
                    [9.95, 69.95, 59.95, 119.95],
                    [59.95, 69.95, 109.95, 119.95],
                    [109.95, 69.95, 159.95, 119.95],
                    [159.95, 69.95, 209.95, 119.95],
                    [9.95, 19.95, 59.95, 69.95],
                    [59.95, 19.95, 109.95, 69.95],
                    [109.95, 19.95, 159.95, 69.95],
                    [159.95, 19.95, 209.95, 69.95],
                ],
                dtype=np.float64,
            ),
        )

    def test_xy_bboxes_is_j_axis_up(self):
        gm = GridMapping.regular(
            size=(2000, 1000), xy_min=(10.0, 20.0), xy_res=0.1, crs=NOT_A_GEO_CRS
        ).derive(is_j_axis_up=True)
        np.testing.assert_almost_equal(
            gm.xy_bboxes, np.array([[9.95, 19.95, 209.95, 119.95]], dtype=np.float64)
        )

        gm = GridMapping.regular(
            size=(2000, 1000),
            xy_min=(10.0, 20.0),
            xy_res=0.1,
            crs=NOT_A_GEO_CRS,
        ).derive(tile_size=500, is_j_axis_up=True)
        np.testing.assert_almost_equal(
            gm.xy_bboxes,
            np.array(
                [
                    [9.95, 19.95, 59.95, 69.95],
                    [59.95, 19.95, 109.95, 69.95],
                    [109.95, 19.95, 159.95, 69.95],
                    [159.95, 19.95, 209.95, 69.95],
                    [9.95, 69.95, 59.95, 119.95],
                    [59.95, 69.95, 109.95, 119.95],
                    [109.95, 69.95, 159.95, 119.95],
                    [159.95, 69.95, 209.95, 119.95],
                ],
                dtype=np.float64,
            ),
        )

    def test_to_coords(self):
        gm = GridMapping.regular(
            size=(10, 6), xy_min=(-2600.0, 1200.0), xy_res=10.0, crs=NOT_A_GEO_CRS
        )

        cv = gm.to_coords(xy_var_names=("x", "y"))
        self._assert_coord_vars(
            cv,
            (10, 6),
            ("x", "y"),
            (-2600.0, -2510.0),
            (1250.0, 1200.0),
            ("x_bnds", "y_bnds"),
            (
                (-2605.0, -2595.0),
                (-2515.0, -2505.0),
            ),
            (
                (1255.0, 1245.0),
                (1205.0, 1195.0),
            ),
        )

    def test_coord_vars_j_axis_up(self):
        gm = GridMapping.regular(
            size=(10, 6), xy_min=(-2600.0, 1200.0), xy_res=10.0, crs=NOT_A_GEO_CRS
        ).derive(is_j_axis_up=True)

        cv = gm.to_coords(xy_var_names=("x", "y"))
        self._assert_coord_vars(
            cv,
            (10, 6),
            ("x", "y"),
            (-2600.0, -2510.0),
            (1200.0, 1250.0),
            ("x_bnds", "y_bnds"),
            (
                (-2605.0, -2595.0),
                (-2515.0, -2505.0),
            ),
            (
                (1195.0, 1205.0),
                (1245.0, 1255.0),
            ),
        )

    def test_coord_vars_antimeridian(self):
        gm = GridMapping.regular(
            size=(10, 10), xy_min=(172.0, 53.0), xy_res=2.0, crs=GEO_CRS
        )

        cv = gm.to_coords(xy_var_names=("lon", "lat"))
        self._assert_coord_vars(
            cv,
            (10, 10),
            ("lon", "lat"),
            (172.0, -170.0),
            (71.0, 53.0),
            ("lon_bnds", "lat_bnds"),
            (
                (171.0, 173.0),
                (-171.0, -169.0),
            ),
            (
                (72.0, 70.0),
                (54.0, 52.0),
            ),
        )

    def _assert_coord_vars(
        self,
        cv,
        size,
        xy_names,
        x_values,
        y_values,
        xy_bnds_names,
        x_bnds_values,
        y_bnds_values,
    ):
        self.assertIsNotNone(cv)
        self.assertIn(xy_names[0], cv)
        self.assertIn(xy_names[1], cv)
        self.assertIn(xy_bnds_names[0], cv)
        self.assertIn(xy_bnds_names[1], cv)

        x = cv[xy_names[0]]
        self.assertEqual((size[0],), x.shape)
        np.testing.assert_almost_equal(x.values[0], np.array(x_values[0]))
        np.testing.assert_almost_equal(x.values[-1], np.array(x_values[-1]))

        y = cv[xy_names[1]]
        self.assertEqual((size[1],), y.shape)
        np.testing.assert_almost_equal(y.values[0], np.array(y_values[0]))
        np.testing.assert_almost_equal(y.values[-1], np.array(y_values[-1]))

        x_bnds = cv[xy_bnds_names[0]]
        self.assertEqual((size[0], 2), x_bnds.shape)
        np.testing.assert_almost_equal(x_bnds.values[0], np.array(x_bnds_values[0]))
        np.testing.assert_almost_equal(x_bnds.values[-1], np.array(x_bnds_values[-1]))

        y_bnds = cv[xy_bnds_names[1]]
        self.assertEqual((size[1], 2), y_bnds.shape)
        np.testing.assert_almost_equal(y_bnds.values[0], y_bnds_values[0])
        np.testing.assert_almost_equal(y_bnds.values[-1], y_bnds_values[-1])

    def test_to_regular(self):
        gm = GridMapping.regular((1000, 1000), (10, 53), 0.01, CRS_WGS84)
        gm_test = gm.to_regular()
        self.assertEqual(gm_test.size, (1000, 1000))
        self.assertEqual(gm_test.tile_size, (1000, 1000))
        self.assertEqual(gm_test.crs, CRS_WGS84)
        self.assertEqual(gm_test.xy_res, (0.01, 0.01))
        self.assertFalse(gm_test.is_j_axis_up)

        gm_test = gm.to_regular(tile_size=500)
        self.assertEqual(gm_test.size, (1000, 1000))
        self.assertEqual(gm_test.tile_size, (500, 500))
        self.assertEqual(gm_test.crs, CRS_WGS84)
        self.assertEqual(gm_test.xy_res, (0.01, 0.01))
        self.assertFalse(gm_test.is_j_axis_up)

        gm_test = gm.to_regular(is_j_axis_up=True)
        self.assertEqual(gm_test.size, (1000, 1000))
        self.assertEqual(gm_test.tile_size, (1000, 1000))
        self.assertEqual(gm_test.crs, CRS_WGS84)
        self.assertEqual(gm_test.xy_res, (0.01, 0.01))
        self.assertTrue(gm_test.is_j_axis_up)

    def test_create_regular_from_bbox(self):
        gm_test = GridMapping.regular_from_bbox(
            [10, 40, 20, 50], 0.1, "EPSG:4326", tile_size=10
        )
        self.assertEqual(gm_test.size, (100, 100))
        self.assertEqual(gm_test.tile_size, (10, 10))
        self.assertEqual(gm_test.crs, CRS_WGS84)
        self.assertEqual(gm_test.xy_res, (0.1, 0.1))
        self.assertFalse(gm_test.is_j_axis_up)



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest

import numpy as np
import xarray as xr

from xcube_resampling.gridmapping import CRS_WGS84, GridMapping
from xcube_resampling.rectify import rectify_dataset

from .sampledata import (
    create_2x2_dataset_with_irregular_coords,
    create_2x2_dataset_with_irregular_coords_antimeridian,
    create_2x2x2_dataset_with_irregular_coords,
    create_4x4_dataset_with_irregular_coords,
)

nan = np.nan


# noinspection PyMethodMayBeStatic
class RectifyDatasetTest(unittest.TestCase):
    def test_rectify_2x2_to_default(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(4, 4), xy_min=(0, 50), xy_res=2, crs=CRS_WGS84
        )
        target_ds = rectify_dataset(
            source_ds,
            target_gm=target_gm,
            interp_methods=0,
            output_indices_names=("indices_lon", "indices_lat"),
        )

        np.testing.assert_almost_equal(
            target_ds.rad.values,
            np.array(
                [
                    [nan, nan, nan, nan],
                    [nan, 1.0, 2.0, nan],
                    [3.0, 3.0, 2.0, nan],
                    [nan, 4.0, nan, nan],
                ],
                dtype=target_ds.rad.dtype,
            ),
        )
        np.testing.assert_almost_equal(
            target_ds.indices_lon.values,
            np.array(
                [
                    [nan, nan, nan, nan],
                    [nan, 0.26086957, 0.60869565, nan],
                    [0.0, 0.42857143, 0.85714286, nan],
                    [nan, 1.0, nan, nan],
                ],
                dtype=target_ds.indices_lon.dtype,
            ),
        )
        np.testing.assert_almost_equal(
            target_ds.indices_lat.values,
            np.array(
                [
                    [nan, nan, nan, nan],
                    [nan, 0.30434783, 0.04347826, nan],
                    [1.0, 0.71428571, 0.42857143, nan],
                    [nan, 1.0, nan, nan],
                ],
                dtype=target_ds.indices_lat.dtype,
            ),
        )

    def test_rectify_2x2_decreasing_x_coords(self):
        source_ds = create_2x2_dataset_with_irregular_coords()
        source_ds = source_ds.isel({"x": slice(None, None, -1)})

        target_gm = GridMapping.regular(
            size=(4, 4), xy_min=(0, 50), xy_res=2, crs=CRS_WGS84
        )
        target_ds = rectify_dataset(
            source_ds,
            target_gm=target_gm,
            interp_methods=0,
            output_indices_names=("indices_lon", "indices_lat"),
        )

        np.testing.assert_almost_equal(
            target_ds.rad.values,
            np.array(
                [
                    [nan, nan, nan, nan],
                    [nan, 1.0, 2.0, nan],
                    [3.0, 4.0, 2.0, nan],
                    [nan, 4.0, nan, nan],
                ],
                dtype=target_ds.rad.dtype,
            ),
        )

    def test_rectify_2x2_to_regular(self):
        source_ds = create_2x2_dataset_with_irregular_coords()
        target_ds = rectify_dataset(source_ds, interp_methods=0)

        np.testing.assert_almost_equal(
            target_ds.rad.values,
            np.array(
                [
                    [nan, nan, nan],
                    [nan, 2.0, nan],
                    [nan, nan, nan],
                ],
                dtype=target_ds.rad.dtype,
            ),
        )

    def test_rectify_2x2x2_to_default(self):
        source_ds = create_2x2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(4, 4), xy_min=(0, 50), xy_res=2, crs=CRS_WGS84
        )
        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)
        self.assertEqual(
            set(source_ds.variables).union(["spatial_ref"]),
            set(target_ds.variables),
        )
        np.testing.assert_almost_equal(
            target_ds.rad.values,
            np.array(
                [
                    [
                        [nan, nan, nan, nan],
                        [nan, 1.0, 2.0, nan],
                        [3.0, 3.0, 2.0, nan],
                        [nan, 4.0, nan, nan],
                    ],
                    [
                        [nan, nan, nan, nan],
                        [nan, 1.0, 2.0, nan],
                        [3.0, 3.0, 2.0, nan],
                        [nan, 4.0, nan, nan],
                    ],
                ],
                dtype=target_ds.rad.dtype,
            ),
        )

    def test_rectify_2x2_to_7x7(self):
        source_ds = create_2x2_dataset_with_irregular_coords()
        # Add offset to "rad" so its values do not lie on a plane
        source_ds["rad"] = source_ds.rad + xr.DataArray(
            np.array([[0.0, 0.0], [0.0, 1.0]]), dims=("y", "x")
        )

        target_gm = GridMapping.regular(
            size=(7, 7), xy_min=(0, 50), xy_res=1.0, crs=CRS_WGS84
        )

        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)

        lon, lat, rad = self._assert_shape_and_dim(target_ds, (7, 7))
        np.testing.assert_almost_equal(lon.values, np.arange(0, 6.1, dtype=lon.dtype))
        np.testing.assert_almost_equal(
            lat.values, np.arange(56, 49.9, -1, dtype=lat.dtype)
        )
        np.testing.assert_almost_equal(
            rad.values,
            np.array(
                [
                    [nan, 1.0, nan, nan, nan, nan, nan],
                    [nan, 1.0, 1.0, nan, nan, nan, nan],
                    [nan, 1.0, 1.0, 1.0, 2.0, nan, nan],
                    [nan, 3.0, 3.0, 1.0, 2.0, 2.0, 2.0],
                    [3.0, 3.0, 3.0, 5.0, 2.0, nan, nan],
                    [nan, 3.0, 5.0, 5.0, nan, nan, nan],
                    [nan, nan, 5.0, nan, nan, nan, nan],
                ],
                dtype=rad.dtype,
            ),
        )

    def test_rectify_2x2_to_7x7_interp_methods_1(self):
        source_ds = create_2x2_dataset_with_irregular_coords()
        # Add offset to "rad" so its values do not lie on a plane
        source_ds["rad"] = source_ds.rad + xr.DataArray(
            np.array([[0.0, 0.0], [0.0, 1.0]]), dims=("y", "x")
        )

        target_gm = GridMapping.regular(
            size=(7, 7), xy_min=(0, 50), xy_res=1.0, crs=CRS_WGS84
        )

        target_ds = rectify_dataset(
            source_ds, target_gm=target_gm, interp_methods="triangular"
        )

        lon, lat, rad = self._assert_shape_and_dim(target_ds, (7, 7))
        np.testing.assert_almost_equal(lon.values, np.arange(0, 6.1, dtype=lon.dtype))
        np.testing.assert_almost_equal(
            lat.values, np.arange(56, 49.9, -1, dtype=lat.dtype)
        )
        np.testing.assert_almost_equal(
            rad.values,
            np.array(
                [
                    [nan, 1.000, nan, nan, nan, nan, nan],
                    [nan, 1.478, 1.391, nan, nan, nan, nan],
                    [nan, 1.957, 1.870, 1.784, 1.697, nan, nan],
                    [nan, 2.435, 2.348, 2.261, 2.174, 2.087, 2.000],
                    [3.000, 3.000, 3.000, 3.000, 3.000, nan, nan],
                    [nan, 4.000, 4.000, 4.000, nan, nan, nan],
                    [nan, nan, 5.000, nan, nan, nan, nan],
                ],
                dtype=rad.dtype,
            ),
            decimal=3,
        )

    def test_rectify_2x2_to_7x7_bilinear_interpol(self):
        source_ds = create_2x2_dataset_with_irregular_coords()
        # Add offset to "rad" so its values do not lie on a plane
        source_ds["rad"] = source_ds.rad + xr.DataArray(
            np.array([[0.0, 0.0], [0.0, 1.0]]), dims=("y", "x")
        )

        target_gm = GridMapping.regular(
            size=(7, 7), xy_min=(0, 50), xy_res=1.0, crs=CRS_WGS84
        )

        target_ds = rectify_dataset(
            source_ds, target_gm=target_gm, interp_methods="bilinear"
        )

        lon, lat, rad = self._assert_shape_and_dim(target_ds, (7, 7))
        np.testing.assert_almost_equal(lon.values, np.arange(0, 6.1, dtype=lon.dtype))
        np.testing.assert_almost_equal(
            lat.values, np.arange(56, 49.9, -1, dtype=lat.dtype)
        )
        np.testing.assert_almost_equal(
            rad.values,
            np.array(
                [
                    [nan, 1.000, nan, nan, nan, nan, nan],
                    [nan, 1.488, 1.410, nan, nan, nan, nan],
                    [nan, 1.994, 1.949, 1.858, 1.722, nan, nan],
                    [nan, 2.520, 2.506, 2.448, 2.344, 2.195, 2.000],
                    [3.000, 3.112, 3.163, 3.153, 3.082, nan, nan],
                    [nan, 4.000, 4.041, 4.020, nan, nan, nan],
                    [nan, nan, 5.000, nan, nan, nan, nan],
                ],
                dtype=rad.dtype,
            ),
            decimal=3,
        )

    def test_rectify_2x2_to_7x7_invalid_interp_methods(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(7, 7), xy_min=(-0.5, 49.5), xy_res=1.0, crs=CRS_WGS84
        )
        with self.assertRaises(NotImplementedError):
            # noinspection PyTypeChecker
            rectify_dataset(source_ds, target_gm=target_gm, interp_methods="cubic")

    def test_rectify_2x2_to_7x7_subset(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(7, 7), xy_min=(2, 51), xy_res=1.0, crs=CRS_WGS84
        )

        target_ds = rectify_dataset(
            source_ds, target_gm=target_gm, interp_methods="nearest"
        )
        lon, lat, rad = self._assert_shape_and_dim(target_ds, (7, 7))
        np.testing.assert_almost_equal(lon.values, np.arange(2, 8.1, dtype=lon.dtype))
        np.testing.assert_almost_equal(
            lat.values, np.arange(57, 50.9, -1, dtype=lat.dtype)
        )
        np.testing.assert_almost_equal(
            rad.values,
            np.array(
                [
                    [nan, nan, nan, nan, nan, nan, nan],
                    [nan, nan, nan, nan, nan, nan, nan],
                    [1.0, nan, nan, nan, nan, nan, nan],
                    [1.0, 1.0, 2.0, nan, nan, nan, nan],
                    [3.0, 1.0, 2.0, 2.0, 2.0, nan, nan],
                    [3.0, 4.0, 2.0, nan, nan, nan, nan],
                    [4.0, 4.0, nan, nan, nan, nan, nan],
                ],
                dtype=rad.dtype,
            ),
        )

    def test_rectify_2x2_to_13x13(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(13, 13), xy_min=(0, 50), xy_res=0.5, crs=CRS_WGS84
        )

        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)

        lon, lat, rad = self._assert_shape_and_dim(target_ds, (13, 13))
        np.testing.assert_almost_equal(lon.values, np.arange(0, 6.1, 0.5, lon.dtype))
        np.testing.assert_almost_equal(lat.values, np.arange(56, 49.9, -0.5, lat.dtype))
        np.testing.assert_almost_equal(rad.values, self.expected_rad_13x13(rad.dtype))

    def test_rectify_2x2_to_13x13_j_axis_up(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(13, 13),
            xy_min=(0, 50),
            xy_res=0.5,
            crs=CRS_WGS84,
            is_j_axis_up=True,
        )

        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)

        lon, lat, rad = self._assert_shape_and_dim(target_ds, (13, 13))
        np.testing.assert_almost_equal(lon.values, np.arange(0, 6.1, 0.5, lon.dtype))
        np.testing.assert_almost_equal(lat.values, np.arange(50, 56.1, 0.5, lat.dtype))
        np.testing.assert_almost_equal(
            rad.values, self.expected_rad_13x13(rad.dtype)[::-1]
        )

    def test_rectify_2x2_to_13x13_j_axis_up_dask_5x5(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(13, 13),
            xy_min=(0, 50),
            xy_res=0.5,
            crs=CRS_WGS84,
            tile_size=5,
            is_j_axis_up=True,
        )

        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)

        lon, lat, rad = self._assert_shape_and_dim(
            target_ds, (13, 13), chunks=((5, 5, 3), (5, 5, 3))
        )
        np.testing.assert_almost_equal(lon.values, np.arange(0, 6.1, 0.5, lon.dtype))
        np.testing.assert_almost_equal(lat.values, np.arange(50, 56.1, 0.5, lat.dtype))
        np.testing.assert_almost_equal(
            rad.values, self.expected_rad_13x13(rad.dtype)[::-1]
        )

    def test_rectify_2x2_to_13x13_dask_7x7(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(13, 13), xy_min=(0, 50), xy_res=0.5, crs=CRS_WGS84, tile_size=7
        )

        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)

        lon, lat, rad = self._assert_shape_and_dim(
            target_ds, (13, 13), chunks=((7, 6), (7, 6))
        )
        np.testing.assert_almost_equal(lon.values, np.arange(0, 6.1, 0.5, lon.dtype))
        np.testing.assert_almost_equal(lat.values, np.arange(56, 49.9, -0.5, lat.dtype))
        np.testing.assert_almost_equal(rad.values, self.expected_rad_13x13(rad.dtype))

    def test_rectify_2x2_to_13x13_dask_5x5(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(13, 13), xy_min=(0, 50), xy_res=0.5, crs=CRS_WGS84, tile_size=5
        )

        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)

        lon, lat, rad = self._assert_shape_and_dim(
            target_ds, (13, 13), chunks=((5, 5, 3), (5, 5, 3))
        )
        np.testing.assert_almost_equal(lon.values, np.arange(0, 6.1, 0.5, lon.dtype))
        np.testing.assert_almost_equal(lat.values, np.arange(56, 49.9, -0.5, lat.dtype))
        np.testing.assert_almost_equal(rad.values, self.expected_rad_13x13(rad.dtype))

    def test_rectify_2x2_to_13x13_dask_3x13(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(13, 13),
            xy_min=(0, 50),
            xy_res=0.5,
            crs=CRS_WGS84,
            tile_size=(3, 13),
        )

        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)

        lon, lat, rad = self._assert_shape_and_dim(
            target_ds, (13, 13), chunks=((13,), (3, 3, 3, 3, 1))
        )
        np.testing.assert_almost_equal(lon.values, np.arange(0, 6.1, 0.5, lon.dtype))
        np.testing.assert_almost_equal(lat.values, np.arange(56, 49.9, -0.5, lat.dtype))
        np.testing.assert_almost_equal(rad.values, self.expected_rad_13x13(rad.dtype))

    def test_rectify_2x2_to_13x13_dask_13x3(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(13, 13),
            xy_min=(0, 50),
            xy_res=0.5,
            crs=CRS_WGS84,
            tile_size=(13, 3),
        )
        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)

        lon, lat, rad = self._assert_shape_and_dim(
            target_ds, (13, 13), chunks=((3, 3, 3, 3, 1), (13,))
        )
        np.testing.assert_almost_equal(lon.values, np.arange(0, 6.1, 0.5, lon.dtype))
        np.testing.assert_almost_equal(lat.values, np.arange(56, 49.9, -0.5, lat.dtype))
        np.testing.assert_almost_equal(rad.values, self.expected_rad_13x13(rad.dtype))

    def test_rectify_2x2_to_13x13_antimeridian(self):
        source_ds = create_2x2_dataset_with_irregular_coords_antimeridian()

        target_gm = GridMapping.regular(
            size=(13, 13), xy_min=(178, 50), xy_res=0.5, crs=CRS_WGS84
        )

        self.assertEqual(True, target_gm.is_lon_360)

        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)

        self.assertIsNotNone(target_ds)
        lon, lat, rad = self._assert_shape_and_dim(target_ds, (13, 13))
        np.testing.assert_almost_equal(
            lon.values,
            np.array(
                [
                    178.0,
                    178.5,
                    179.0,
                    179.5,
                    180.0,
                    -179.5,
                    -179.0,
                    -178.5,
                    -178.0,
                    -177.5,
                    -177.0,
                    -176.5,
                    -176.0,
                ],
                dtype=lon.dtype,
            ),
        )
        np.testing.assert_almost_equal(lat.values, np.arange(56, 49.9, -0.5, lat.dtype))
        np.testing.assert_almost_equal(rad.values, self.expected_rad_13x13(rad.dtype))

    def test_rectify_2x2_to_13x13_none(self):
        source_ds = create_2x2_dataset_with_irregular_coords()

        target_gm = GridMapping.regular(
            size=(13, 13), xy_min=(10.0, 50.0), xy_res=0.5, crs=CRS_WGS84
        )
        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_array_equal(
            np.isnan(target_ds.rad), np.ones_like(target_ds.rad, dtype=bool)
        )

        target_gm = GridMapping.regular(
            size=(13, 13), xy_min=(-10.0, 50.0), xy_res=0.5, crs=CRS_WGS84
        )
        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_array_equal(
            np.isnan(target_ds.rad), np.ones_like(target_ds.rad, dtype=bool)
        )

        target_gm = GridMapping.regular(
            size=(13, 13), xy_min=(0.0, 58.0), xy_res=0.5, crs=CRS_WGS84
        )
        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_array_equal(
            np.isnan(target_ds.rad), np.ones_like(target_ds.rad, dtype=bool)
        )

        target_gm = GridMapping.regular(
            size=(13, 13), xy_min=(0.0, 42.0), xy_res=0.5, crs=CRS_WGS84
        )
        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_array_equal(
            np.isnan(target_ds.rad), np.ones_like(target_ds.rad, dtype=bool)
        )

    def test_rectify_different_crs(self):
        source_ds = create_4x4_dataset_with_irregular_coords()
        target_gm = GridMapping.regular(
            size=(3, 3), xy_min=(3650000, 3250000), xy_res=100000, crs="epsg:3035"
        )
        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)

        np.testing.assert_almost_equal(
            target_ds.x.values, np.array([3650000.0, 3750000.0, 3850000.0])
        )
        np.testing.assert_almost_equal(
            target_ds.y.values, np.array([3450000.0, 3350000.0, 3250000.0])
        )
        np.testing.assert_almost_equal(
            target_ds.rad.values,
            np.array(
                [[10.0, 6.0, 3.0], [10.0, 7.0, 3.0], [11.0, 11.0, 8.0]],
                dtype=target_ds.rad.dtype,
            ),
        )

    def test_rectify_different_crs_dask(self):
        source_ds = create_4x4_dataset_with_irregular_coords()
        source_ds = source_ds.chunk(dict(y=2, x=2))
        target_gm = GridMapping.regular(
            size=(3, 3), xy_min=(3650000, 3250000), xy_res=100000, crs="epsg:3035"
        )
        target_ds = rectify_dataset(source_ds, target_gm=target_gm, interp_methods=0)
        np.testing.assert_almost_equal(
            target_ds.x.values, np.array([3650000.0, 3750000.0, 3850000.0])
        )
        np.testing.assert_almost_equal(
            target_ds.y.values, np.array([3450000.0, 3350000.0, 3250000.0])
        )
        np.testing.assert_almost_equal(
            target_ds.rad.values,
            np.array(
                [[10.0, 6.0, 3.0], [10.0, 7.0, 3.0], [11.0, 11.0, 8.0]],
                dtype=target_ds.rad.dtype,
            ),
        )

    def _assert_shape_and_dim(
        self, target_ds, size, chunks=None, var_names=("rad",)
    ) -> tuple[xr.DataArray, ...]:
        w, h = size

        self.assertIn("lon", target_ds)
        lon = target_ds["lon"]
        self.assertEqual((w,), lon.shape)
        self.assertEqual(("lon",), lon.dims)

        self.assertIn("lat", target_ds)
        lat = target_ds["lat"]
        self.assertEqual((h,), lat.shape)
        self.assertEqual(("lat",), lat.dims)

        # noinspection PyShadowingBuiltins
        vars = []
        for var_name in var_names:
            self.assertIn(var_name, target_ds)
            var = target_ds[var_name]
            self.assertEqual((h, w), var.shape)
            self.assertEqual(("lat", "lon"), var.dims)
            self.assertEqual(chunks, var.chunks)
            vars.append(var)

        return lon, lat, *vars

    def expected_rad_13x13(self, dtype):
        return np.array(
            [
                [nan, nan, 1.0, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan],
                [nan, nan, 1.0, 1.0, nan, nan, nan, nan, nan, nan, nan, nan, nan],
                [nan, nan, 1.0, 1.0, 1.0, 1.0, nan, nan, nan, nan, nan, nan, nan],
                [nan, nan, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, nan, nan, nan, nan, nan],
                [nan, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, nan, nan, nan, nan],
                [nan, 3.0, 3.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, nan, nan],
                [nan, 3.0, 3.0, 3.0, 3.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
                [nan, 3.0, 3.0, 3.0, 3.0, 3.0, 1.0, 2.0, 2.0, 2.0, 2.0, nan, nan],
                [3.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0, 2.0, 2.0, 2.0, nan, nan, nan],
                [nan, 3.0, 3.0, 3.0, 4.0, 4.0, 4.0, 4.0, 2.0, nan, nan, nan, nan],
                [nan, nan, 3.0, 4.0, 4.0, 4.0, 4.0, nan, nan, nan, nan, nan, nan],
                [nan, nan, nan, 4.0, 4.0, 4.0, nan, nan, nan, nan, nan, nan, nan],
                [nan, nan, nan, nan, 4.0, nan, nan, nan, nan, nan, nan, nan, nan],
            ],
            dtype=dtype,
        )
